#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  autoenchoder_V0.0.py
#
#  Copyright 2018  <@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#
import operator
from .autoenc_solo4 import autoencoder
from scipy.stats import pearsonr
import parser

letters={'CYS': 'C', 'ASP': 'D', 'SER': 'S', 'GLN': 'Q', 'LYS': 'K', 'ASN': 'N', 'PRO': 'P', 'THR': 'T', 'PHE': 'F', 'ALA': 'A', 'HIS': 'H', 'GLY': 'G', 'ILE': 'I', 'LEU': 'L', 'ARG': 'R', 'TRP': 'W', 'VAL': 'V', 'GLU': 'E', 'TYR': 'Y', 'MET': 'M'}
from .parser import protein_class, residue_class
import torch
import torch.autograd as autograd
import torch.nn as nn

torch.manual_seed(1)
EPOCH=10
import os
import numpy as np
from sklearn.preprocessing import MinMaxScaler,StandardScaler

def run_rsa(folder='datasets/rereferenced_nmr/',outdir='rsa_vals/'):
	rsabin='./pdbasa'
	#for i in os.listdir(folder):
	#	os.system(rsabin+' '+folder+i+' '+outdir+i)
	## parse ##
	diz={}
	v=[]
	seqs={}
	for i in os.listdir(outdir):
		nome=i.split('.')[0]
		diz[nome]=[]
		resold=1
		tmp=[]
		cont=0
		seq=''
		for j in open(outdir+i,'r').readlines():

			if j.strip()=='':
				continue
			if not 'ATOM' == j[0:4]:
				continue
			res=j[22:26]
			rsa=j[60:66]
			try:
				aa=letters[j[17:20]]
			except:
				continue
			v+=[float(rsa)]
			if res!=resold and cont!=0:

				diz[nome]+=[np.mean(tmp)]
				seq+=aa
				tmp=[]
			tmp+=[-float(rsa)]
			resold=res
			cont+=1

		diz[nome]+=[np.mean(tmp)]
		seq+=aa
		assert len(seq)==len(diz[nome])
		seqs[nome]=seq
	v=np.array(v).reshape((-1,1))
	scaler=MinMaxScaler()
	scaler.fit(v)

	#print diz[i]
	for i in list(diz.keys()):
		try:
			diz[i]=scaler.transform(np.array(diz[i]).reshape((-1,1))).reshape((-1))
		except:
			diz[i]=[-10]*len(seqs[i])
	return diz,seqs
def leggifasta(database): #legge un file fasta e lo converte in un dizionario
	f=open(database)
	uniprot=f.readlines()
	f.close()
	dizio={}
	for i in uniprot:

		if i[0]=='>':
				if '|' in i:

					uniprotid=i.strip('>\n').split('|')[1]
				else:
					uniprotid=i.strip('>\n')
				dizio[uniprotid]=''
		else:
			dizio[uniprotid]=dizio[uniprotid]+i.strip('\n')
	return dizio
def parse_uniprot_annotations(fil='dataset/all_seqs_dnabinding_uniprot.annotations'):
	diz={}
	salta=0
	for i in open(fil).readlines():
		a=i.split()
		try:
			bind_start=int(a[3])-1
			bind_end=int(a[4])
		except:
			salta+=1
			continue
		seq=a[1]
		p1=[0]*bind_start
		p2=[1]*(bind_end-bind_start)
		p3=[0]*(len(seq)-bind_end)
		p=p1+p2+p3
		assert len(p)==len(seq)
		diz[a[0]]=(seq,p)
	#print 'got ',len(diz),'predictions. Saltate',salta
	return diz
def getScoresSVR(pred, real, threshold,FULL_SCORES=True):
	import math
	import numpy as np
	if len(pred) != len(real):
		raise Exception("ERROR: input vectors have differente len!")
	i = 0
	confusionMatrix = {}
	confusionMatrix["TP"] = confusionMatrix.get("TP", 0)
	confusionMatrix["FP"] = confusionMatrix.get("FP", 0)
	confusionMatrix["FN"] = confusionMatrix.get("FN", 0)
	confusionMatrix["TN"] = confusionMatrix.get("TN", 0)
	while i < len(real):
		if float(pred[i])<=threshold and (int(real[i])==0):
			confusionMatrix["TN"] = confusionMatrix.get("TN", 0) + 1
		if float(pred[i])<=threshold and int(real[i])==1:
			confusionMatrix["FN"] = confusionMatrix.get("FN", 0) + 1
		if float(pred[i])>=threshold and int(real[i])==1:
			confusionMatrix["TP"] = confusionMatrix.get("TP", 0) + 1
		if float(pred[i])>=threshold and int(real[i])==0:
			confusionMatrix["FP"] = confusionMatrix.get("FP", 0) + 1
		i += 1
	#print confusionMatrix
	if FULL_SCORES:

		#print "--------------------------------------------"
		#print confusionMatrix["TN"],confusionMatrix["FN"],confusionMatrix["TP"],confusionMatrix["FP"]
		#print "          | SSBOND            | FREE             |"
		#print "predBond  | TP: %d (%2.2f%%)  | FP: %d (%2.2f%%) |" % (confusionMatrix["TP"], (confusionMatrix["TP"]/float(confusionMatrix["TP"]+confusionMatrix["FN"]))*100, confusionMatrix["FP"], (confusionMatrix["FP"]/float(confusionMatrix["FP"]+confusionMatrix["TN"]))*100 )
		#print "predFree  | FN: %d (%2.2f%%)  | TN: %d (%2.2f%%) |" % (confusionMatrix["FN"],(confusionMatrix["FN"]/float(confusionMatrix["TP"]+confusionMatrix["FN"]))*100, confusionMatrix["TN"], (confusionMatrix["TN"]/float(confusionMatrix["FP"]+confusionMatrix["TN"]))*100)
		sen = (confusionMatrix["TP"]/float((confusionMatrix["TP"] + confusionMatrix["FN"])))
		spe = (confusionMatrix["TN"]/float((confusionMatrix["TN"] + confusionMatrix["FP"])))
		acc =  (confusionMatrix["TP"] + confusionMatrix["TN"])/float((sum(confusionMatrix.values())))
		bac = (0.5*((confusionMatrix["TP"]/float((confusionMatrix["TP"] + confusionMatrix["FN"])))+(confusionMatrix["TN"]/float((confusionMatrix["TN"] + confusionMatrix["FP"])))))
		inf =((confusionMatrix["TP"]/float((confusionMatrix["TP"] + confusionMatrix["FN"])))+(confusionMatrix["TN"]/float((confusionMatrix["TN"] + confusionMatrix["FN"])))-1.0)
		pre =(confusionMatrix["TP"]/float((confusionMatrix["TP"] + confusionMatrix["FP"])))
		mcc =	( ((confusionMatrix["TP"] * confusionMatrix["TN"])-(confusionMatrix["FN"] * confusionMatrix["FP"])) / math.sqrt((confusionMatrix["TP"]+confusionMatrix["FP"])*(confusionMatrix["TP"]+confusionMatrix["FN"])*(confusionMatrix["TN"]+confusionMatrix["FP"])*(confusionMatrix["TN"]+confusionMatrix["FN"])) )

	#print real
	#print pred
	from sklearn.metrics import roc_auc_score
	aucScore = roc_auc_score(real, pred)

	#print "AUC = %3.3f" % aucScore
	#print "--------------------------------------------"
	if FULL_SCORES:
		return sen,spe,acc,pre,mcc,aucScore
	else:
		return aucScore
def run_ring(pdb_folder='pdb_interacting/'):
	############ you need to export VICTOR_ROOT=/home/gabriele/HD1/dna_binding_prediction/ring/ ####
	def parse_ring(fil,tmp_fold='tmp_ringfiles/'):
		fil=tmp_fold+fil
		diz={}
		for i in open(fil).readlines():
			a=i.split()
			if ':_:DA' in a[0] or ':_:DT' in a[0] or ':_:DC' in a[0] or ':_:DG' in a[0]:
				first_dna=True
			else:
				first_dna=False
			if ':_:DA' in a[2] or ':_:DT' in a[2] or ':_:DC' in a[2] or ':_:DG' in a[2]:
				second_dna=True
			else:
				second_dna=False
			if first_dna and not second_dna:
				resi=a[2].split(':')
				if resi[3] not in letters:
					print('MISSING AMINOACID')
					rname='X'
				else:
					rname=letters[resi[3]]
				if resi[0] not in diz:
					diz[resi[0]]={}
				diz[resi[0]][(int(resi[1]),rname)]=1
			if not first_dna and second_dna:
				resi=a[0].split(':')
				if resi[3] not in letters:
					print('MISSING AMINOACID')
					rname='X'
				else:
					rname=letters[resi[3]]
				if resi[0] not in diz:
					diz[resi[0]]={}
				diz[resi[0]][(int(resi[1]),rname)]=1
		return diz
	def parse_pdb(fil,tmp_fold='pdb_interacting/'):
		fil=tmp_fold+fil
		diz={}
		seq={}
		for i in open(fil).readlines():
			if i.strip()=='':
				continue
			a=i.split()
			if not a[0]=='ATOM':
				continue
			if 'DA' in a[3] or 'DT' in a[3] or 'DC' in a[3] or 'DG' in a[3]: #dna
				continue
			chain=i[21]#a[4]
			resnum=int(i[22:26])
			if chain not in seq:
				seq[chain]={}
			'''
			try:
				int(a[5])
			except:
				return 'error'
			'''

			if a[3] not in letters:
				print('MISSING AMINOACID')
				rname='X'
			else:
				rname=letters[a[3]]
			if (int(resnum),rname) not in seq[chain]:
						seq[chain][(int(resnum),rname)]=0
			#except:
			#	return 'error'
			'''
			except:
				i=i[:22]+' '+i[22:] ### ma porcodio...se i residui vanno sopra 999 il pdb si dimentica di aggiungere una colonna
				a=i.split()
				if not a[0]=='ATOM':
					continue
				if 'DA' in a[3] or 'DT' in a[3] or 'DC' in a[3] or 'DG' in a[3]: #dna
					continue
				chain=a[4]
				if not seq.has_key(chain):
					seq[chain]={}

				if not seq[chain].has_key((int(a[5]),letters[a[3]])):
					seq[chain][(int(a[5]),letters[a[3]])]=0
				#raw_input()
			'''
		return seq
	ring_bin='ring/bin/Ring'
	db={}
	cont=0
	for pdb in os.listdir(pdb_folder)[:]:
		if not '.pdb' in pdb:
			continue
		#if pdb.replace('.pdb','.edges') in os.listdir('tmp_ringfiles'):
		#	continue
		print('starting ',pdb)
		os.system(ring_bin+' -i '+pdb_folder+pdb+' -t 3 --all -E tmp_ringfiles/'+pdb.replace('.pdb','.edges')+' > /dev/null')

		diz=parse_ring(pdb.replace('.pdb','.edges'))
		pdb_diz=parse_pdb(pdb)
		if pdb_diz=='error':
			print('salto')
			continue
		### chain selection policy i take the one with the greatest number of interactions ###
		bestchain=0
		for i in list(diz.keys()):
			if len(diz[i])>bestchain:
				chain=i
				bestchain=len(diz[i])
		if bestchain==0:
			print('NO DNA INTERACTIONS IN',pdb.replace('.pdb',''))
			continue
		if chain not in pdb_diz:
			print('salto')
			continue
		for i in list(diz[chain].keys()):
			pdb_diz[chain][i]=1

		seq=''
		vals=[]
		for i in sorted(pdb_diz[chain].keys()):
			seq+=i[1]
			vals+=[pdb_diz[chain][i]]
		db[pdb.replace('.pdb','')]=(seq,vals)


	return db

def aucNth(yp, y, N=50):
	assert len(y) == len(yp)
	assert len(y) > 1
	from sklearn.metrics import roc_curve, auc
	fpr, tpr, thresholds = roc_curve(y, yp)
	negatives = y.count(0)
	assert N < negatives
	perc = N / float(negatives)
	#print perc
	fpr1k = []
	tpr1k = []
	i = 0
	while i < len(fpr):
		if fpr[i] > perc:
			break
		fpr1k.append(fpr[i])
		tpr1k.append(tpr[i])
		i+=1
	assert len(fpr1k) > 1
	#print fpr1k, tpr1k
	aucScore = auc(fpr1k, tpr1k) / perc
	#print ">>AUC%d: %3.3f" %(N,aucScore)
	return aucScore

class nn_class_autoenc(nn.Module):
		def __init__(self, n_features,hidden_dim=10,cuda=False):
			super(nn_class_autoenc, self).__init__()

			self.encoder = nn.Sequential(
				nn.Linear(n_features, hidden_dim),
				nn.ReLU(),
				nn.Linear(hidden_dim, hidden_dim),
				nn.ReLU(),
				nn.Linear( hidden_dim, 1),
				nn.Sigmoid()
				)
			self.decoder = nn.Sequential(
				nn.Linear( 1,hidden_dim),
				nn.ReLU(),
				nn.Linear(hidden_dim, hidden_dim),
				nn.ReLU(),
				nn.Linear(hidden_dim,n_features),
				nn.Sigmoid()
			)

		def forward(self, x):
				encoded = self.encoder(x)
				decoded = self.decoder(encoded)
				return encoded, decoded
		def encode_x(self,x):
			return self.encoder(x)
		def decode_x(self,x):
			return self.decoder(x)

class autoencoder():
	def __init__(self, hidden_dim=100,cuda=False):
		self.hidden_dim=hidden_dim
		self.cuda=cuda
	def fit(self,x,batch=100,EPOCH=10,LR=0.01):
		nn_auto=nn_class_autoenc(len(x[0]),hidden_dim=self.hidden_dim)
		optimizer = torch.optim.Adam(nn_auto.parameters())
		loss_func = nn.MSELoss()
		#plt.ion()
		#fig, ax = plt.subplots(nrows=1, ncols=1)
		for epoch in range(EPOCH):
			decoded_all=[]
			x_all=[]
			for i in range(0,len(x),batch):
				if i+batch>len(x):
					end=len(x)
				else:
					end=len(x)+i
				if self.cuda:
					b_x = autograd.Variable(torch.Tensor(x[i:end])).cuda()
					b_y = autograd.Variable(torch.Tensor(x[i:end])).cuda()
				else:
					b_x = autograd.Variable(torch.Tensor(x[i:end]))
					b_y = autograd.Variable(torch.Tensor(x[i:end]))

				encoded, decoded = nn_auto(b_x)
				decoded_all+=list(decoded.data.numpy())
				x_all+=list(x[i:end])
				loss = loss_func(decoded[:,:4], b_y[:,:4])
				optimizer.zero_grad()
				loss.backward()



				#ax.plot([np.array(x_all)[:,0],np.array(x_all)[:,1]],'ro')


				optimizer.step()
			#print epoch,loss.data.numpy()[0]
			'''
			ax.clear()
			a=np.array(decoded_all)[0:100,0]
			b=np.array(decoded_all)[0:100,1]

			#assert len(a)==len(b)

			ax.scatter(a,b,alpha=1)
			ax.scatter(np.array(x_all)[:,0],np.array(x_all)[:,1],color='r',alpha=0.01)
			plt.draw(); plt.pause(0.05)
			print loss.data

			ax.clear()
			ax.plot([np.array(decoded_all)[0:100,0],np.array(decoded_all)[0:100,1]],'bo')
			ax.plot([np.array(x_all)[:,0],np.array(x_all)[:,1]],'ro')
			plt.draw(); plt.pause(0.05)
			'''
		self.model=  nn_auto
	def encode(self,x):

		if self.cuda:
			b_x = autograd.Variable(torch.Tensor(x)).cuda()
			b_y = autograd.Variable(torch.Tensor(x)).cuda()
		else:
			b_x = autograd.Variable(torch.Tensor(x))
			b_y = autograd.Variable(torch.Tensor(x))
		y_b=self.model.encode_x(b_x)
		#print len(y_b)
		return y_b.data.numpy().flatten()
	def decode(self,x):

		if self.cuda:
			b_x = autograd.Variable(torch.Tensor(x)).cuda()
			b_y = autograd.Variable(torch.Tensor(x)).cuda()
		else:
			b_x = autograd.Variable(torch.Tensor(x))
			b_y = autograd.Variable(torch.Tensor(x))
		y_b=self.model.decode_x(b_x)

		return y_b.data.numpy()

class autoencoder_transformation():
	def __init__(self,chemical_shifts_diz):
		self.scalers={}
		self.autoen={}
		self.shifts={}
		self.second_scaler={}
		self.third_scaler={}
		self.inverse={}
		self.sec_struct_scaler={}
		#################################
		#### change here the scalers ####
		#################################
		# remember the first one MUST mantain the shape --> don't use standard scaler or you will get random fitting
		self.first_scaler_class=MinMaxScaler
		self.second_scaler_class=StandardScaler
		self.third_scaler_class=MinMaxScaler
		self.sec_struct_scaler_class=None#secondary_structure_scaler
		self.boundaries={}
		self.shifts=chemical_shifts_diz
		#################################
	def fit(self,training_folder,remove_outlayers=True): #training _folder contains the nef files for training
		diz_base={}

		diz_base=parser.parse_train(training_folder)
		predictor={}
		scalers={}
		diz={}
		for prot in list(diz_base.keys()):
			for resi in diz_base[prot].resi:
				if resi.resname not in diz:
					diz[resi.resname]=[]

				diz[resi.resname]+=[resi]
		count_available_chemical_shft=False
		if count_available_chemical_shft:
			for i in list(diz.keys()):
				conteggio={}
				print(i)
				#if i=='G' or i=='P':
				#	continue
				for r in diz[i]:
					for k in list(r.atom.keys()):
						if k not in conteggio:
							conteggio[k]=0

						conteggio[k]+=1
				for k in sorted(list(conteggio.items()), key=operator.itemgetter(1)):
					print('\t',k[0],k[1])
		self.average={}
		#print sorted(diz.keys())
		pearson_resu={}
		for aa in sorted(diz.keys())[:]:
			print('starting',aa)
			ss=[]
			X=[]
			if aa not in self.average:
				self.average[aa]={}
			for entry in diz[aa][:]:
				xt=[]
				buona=True

				for i in self.shifts[aa]:

					if i not in entry.atom:
						buona=False

						break
				if buona:

					for i in self.shifts[aa]:
						xt+=[entry.atom[i]]
						if i not in self.average[aa]:
							self.average[aa][i]=[]
						self.average[aa][i]+=[entry.atom[i]]
					X+=[xt]
					ss+=[entry.ss]
			for i in list(self.average[aa].keys()):
				self.average[aa][i]=np.mean(self.average[aa][i])
			scaler=self.first_scaler_class()
			if X==[]:
				print('## ERROR NOT ENOUGH DATA TO TRAIN THE DEFINED MODEL ##')
				print('try to reduce the number/type of atom taken into consideration for amino acid',aa)
				print('Another possibility is that you mispelled an atom code. They need to be the same of the ones in the training files(sources/datasets/rereferenced_nmr/)')
				assert False
			X=np.array(X)
			if remove_outlayers==True:
				perc=[]
				Xnew=[]
				ssnew=[]
				goodpos=[]
				fin=np.ones(len(X),dtype=bool)

				for i in range(len(self.shifts[aa])):

					up=np.percentile(X[:,i],99)
					down=np.percentile(X[:,i],1)
					low_values_flags = X[:,i] > down
					high_values_flags = X[:,i] < up
					fin=np.logical_and(np.logical_and(low_values_flags,high_values_flags),fin)

				for i in range(len(X)):
					if fin[i]==True:
						Xnew+=[X[i]]
						ssnew+=[ss[i]]
					else:
						pass
				print('removed',len(X)-len(Xnew),'outlaiers starting from',len(X),'points')
				X=np.array(Xnew)
				#print X.shape

				ss=ssnew

			X=scaler.fit_transform(X)
			self.scalers[aa]=scaler
			self.autoen[aa]=autoencoder(hidden_dim=50)
			self.autoen[aa].fit(X,batch=100,EPOCH=10,LR=0.01)
			yp=self.autoen[aa].encode(X)
			back_to_x=self.autoen[aa].decode(yp.reshape(-1,1))
			pear=True

			if pear:
				f=open('pear','a')
				for sh in range(len(self.shifts[aa])-1,-1,-1):
					if self.shifts[aa][sh] not in pearson_resu:
						pearson_resu[self.shifts[aa][sh]]=[]
						pearson_resu[self.shifts[aa][sh]]+=[pearsonr(X[:,sh],back_to_x[:,sh])[0]]
					f.write(aa+'\t'+self.shifts[aa][sh]+'\t'+str(np.round(pearsonr(X[:,sh],back_to_x[:,sh])[0],2))+'\n')
					print('\tpearosn for arom:',self.shifts[aa][sh], pearsonr(X[:,sh],back_to_x[:,sh])[0])
					#plt.scatter(X[:,sh],back_to_x[:,sh])
					#plt.show()
			colors={'H':'g','E':'r','C':'b','-':'k'}


			yp_original=yp[:]
			vals=np.percentile(yp,10)
			sec=[]
			for i in range(len(yp)):
				if yp[i]<vals:
					sec+=ss[i]
			print('\tH/E rate:',float(sec.count('H'))/(float(sec.count('E')+1)))
			if sec.count('H')>sec.count('E'):
				self.inverse[aa]=False
			else:
				self.inverse[aa]=True

			if self.inverse[aa]==True:
				yp= np.array([ -x for x in yp])

			if self.second_scaler_class!=None:
				self.second_scaler[aa]=self.second_scaler_class()
				yp=self.second_scaler[aa].fit_transform(yp.reshape(-1,1)).reshape(-1)
			if self.third_scaler_class!=None:
				self.third_scaler[aa]=self.third_scaler_class()
				yp=self.third_scaler[aa].fit_transform(yp.reshape(-1,1)).reshape(-1,1)
			'''
			n, bins, patches = plt.hist(yp, 50, histtype='stepfilled')
			plt.setp(patches, 'facecolor', 'blue', 'alpha', 0.3)
			#plt.setp(patches, 'facecolor', 'alpha', 0.3)
			plt.show()
			'''
			####inverse:
			'''
			if self.third_scaler_class!=None:
				yp=self.third_scaler[aa].inverse_transform(yp.reshape(-1,1)).reshape(-1,1)
			if self.second_scaler_class!=None:
				yp=self.second_scaler[aa].inverse_transform(yp.reshape(-1,1)).reshape(-1,1)
			if self.inverse[aa]==True:
				yp= np.array([ -x for x in yp])
			'''
			if self.sec_struct_scaler_class!=None:
				self.sec_struct_scaler[aa]=self.sec_struct_scaler_class()
				self.sec_struct_scaler[aa].fit(yp,ss)
				if self.sec_struct_scaler[aa]!=None:
					yp=self.sec_struct_scaler[aa].transform(np.array(yp).reshape(-1,1)).reshape(-1)
				#xs = np.linspace(0,1,200)
				#yxs=self.sec_struct_scaler[aa].transform(xs.reshape(-1,1)).reshape(-1)
				#plt.plot(xs,yxs)
			plot_single_atom=False
			if plot_single_atom:
				raw_cs=X
				raw_cs=np.array(raw_cs)
				for i in range(len(self.shifts[aa])):
					plt.title('aa '+aa+' atom: '+self.shifts[aa][i])
					plt.xlabel('Original CS (scaled from 0 to 1)')
					plt.ylabel('NN-Transformed CS')
					plt.xlim(0,1)
					plt.ylim(0,1)
					for j in range(len(raw_cs)):
						plt.plot([raw_cs[j][i]],[yp[j]],color=colors[ss[j]],marker=',',alpha=0.5)

					plt.clf()
					#plt.show()

	def transform(self,protein): ##takes a protein instance
		res=[]
		for i in protein.resi:
			xt=[]
			buona=True
			aa=i.resname
			if aa=='X':
				res+=[-10]
				continue
			for atom in self.shifts[aa]:
				if atom not in i.atom:
					#i.atom[atom]=self.average[aa][atom]
					buona=False
					break
			if buona:
				for atom in self.shifts[aa]:
					xt+=[i.atom[atom]]
				xt=self.scalers[aa].transform([xt])
				yp=self.autoen[aa].encode(xt)
				if self.inverse[aa]==True:
					yp= np.array([ -x for x in yp])
				if self.second_scaler_class!=None:
					yp=self.second_scaler[aa].transform([yp])
				if self.third_scaler_class!=None:
					yp=self.third_scaler[aa].transform(yp)
				if self.sec_struct_scaler_class!=None:
					yp=self.sec_struct_scaler[aa].transform(yp)
				res+=list(yp[0]) ###	BUG QUI, VIENE DIVERSO SE USI SCALERS OPPURE NO, TIPO UNA LISTA IN PIU
			else:

				res+=[-10]
		#print res
		return np.array(res)
	def inverse_transform(self,yp,seq): ##takes a protein instance
		yp=np.array(yp)
		fin=[]
		prot=protein_class()
		for ind in range(len(seq)):
			resi=residue_class()
			aa=seq[ind]
			val=[[yp[ind]]]
			if self.sec_struct_scaler_class!=None and self.sec_struct_scaler[aa]!=None:
				val=self.sec_struct_scaler[aa].inverse_transform(val)
			if self.third_scaler_class!=None:
				val=self.third_scaler[aa].inverse_transform(val)
			if self.second_scaler_class!=None:

				val=self.second_scaler[aa].inverse_transform(val)
			if self.inverse[aa]==True:
				val= [[-val[0][0]]]
			xt=self.autoen[aa].decode(val)[0]  ###ANCHE QUI, VEDI SOPRA
			xt=self.scalers[aa].inverse_transform([xt])[0]
			atoms={}
			assert len(xt)==len(self.shifts[aa])
			for i in range(len(self.shifts[aa])):
				atoms[self.shifts[aa][i]]=float(xt[i])
			resi.resname=aa
			resi.atom=atoms
			prot.resi+=[resi]
			prot.seq=seq

			#xt=list(xt)

			#fin+=[xt]

		return prot

def main(args):
	a=autoencoder_transformation()
	a.fit(training_folder='datasets/rereferenced_nmr')
	torch.save(a,'../marshalled/custom_model_10epocs.mtorch')
	#a.transform('../training_dataset_example/bmr4031_1.nef')

if __name__ == '__main__':
	import sys
	sys.exit(main(sys.argv))



import time
import os
import json
import requests as r
from io import StringIO
import pandas as pd
import numpy as np
import time,sys,csv

class B2bIo:
  def __init__(self, gapCode='-'):
      self.gapCode = gapCode

  references = [] # Defined in subclass for 'traditional' dynaMine file format output

  #
  # File reading
  #

  def readFasta(self, fileName, fileString=None):

    """
    Reads a FASTA file - from dynaMine.predictor.parsers import DynaMineFileParsers

    :param fileName: Path to the FASTA file
    :return: List of (seqId,sequenceString) tuples
    """

    if fileName:
      print("Reading FASTA file {}...".format(fileName))

      # Bypassing FC; much faster to do like this...
      fin = open(fileName)
      lines = fin.readlines()
      fin.close()
    else:
      lines = fileString.split("\n") # WV this might not always work

    # Quick FASTA read, handles multiple line sequences
    seqs = []
    seqId = None
    sequence = ""
    for line in lines:
      if line.startswith(">"):
        # Consolidate
        if seqId:
          seqs.append((seqId, sequence.upper()))
        cols = line.split()
        seqId = self.convertSeqId(cols[0][1:])
        sequence = ""
      elif line.strip():
        sequence += line.strip().replace(' ', '')
    if seqId:
      seqs.append((seqId, sequence.upper()))

    return seqs

  def writeFasta(self, fastaFileName, sequences):

    fout = open(fastaFileName, 'w')

    for (seqId, sequence) in sequences:
      fout.write(">{} {}\n".format(seqId, seqId))
      fout.write("{}\n\n".format(sequence))

    fout.close()

  def convertSeqId(self,seqId):

    # Possibility to subclass for nicer identifiers (e.g. uniprot default ones)

    return seqId

  #
  # Wrapper defs for predictions
  #

  def predictFromFasta(self, fastaFile, testMode=False):

    """
    Get predictions in relation to a FASTA file. Prediction type defined by subclass
    :param fastaFile: Path to a FASTA file
    :return: Predictions for a subclass (as stored in allPredictions)
    """

    print("Reading input fasta...", end=' ')
    seqs = self.readFasta(fastaFile)

    if testMode:
       seqs = seqs[:5]

    self.predictSeqs(seqs)

    return self.allPredictions

  def predictSeqs(self,seqs):

    # Defined in subclass
    pass

  #
  # Writing 'classic' DynaMine format files
  #

  def writeAllPredictions(self, identifier, outputDir='dynaMineResults'):

    """
    Writes out all available predictions with identifier as base name into outputDir.
    :param identifier: Base name for the output files
    :keyword outputDir: Default is dynaMineResults/, can be reset to wherever
    :return: None
    """

    if not os.path.exists(outputDir):
      os.makedirs(outputDir)

    seqIds = list(self.allPredictions.keys())
    predictionTypes = list(self.allPredictions[seqIds[0]].keys())
    for predictionType in predictionTypes:
      filePath = os.path.join(outputDir,"{}_{}.pred".format(identifier,predictionType))
      self.writePredictionFile(filePath,predictionType)

  def writePredictionFile(self, outFile, predictionType):

    """
    Write out the predictions to the 'old' DynaMine format.
    Assumes that self.allPredictions is set by a subclass!
    :param outFile: Name of the output file
    :param predictionType: Prediction type (backbone, coil, earlyFolding, ...)
    :return: None
    """

    seqIds = list(self.allPredictions.keys())
    seqIds.sort()

    date = time.strftime("%Y.%m.%d_%H.%M.%S")

    referencesText = self.getPredFileReferences()
    infoText = self.getPredFileInfoText(predictionType)

    (pathName, baseName) = os.path.split(outFile)
    if pathName and not os.path.exists(pathName):
      os.makedirs(pathName)

    fout = open(outFile, 'w')

    for seqId in seqIds:
      fout.write(self.getPredFileHeader().format(infoText, seqId, date, referencesText))

      # TODO: Enable writing out discrete value (or category for backbone?)
      for predInfo in self.allPredictions[seqId][predictionType]:
        fout.write("{}	{:7.3f}\n".format(predInfo[0], predInfo[1]))

    fout.close()

  def getPredFileInfoText(self,predictionType):

    predFileInfoText = """* {:<48s} *\n""".format("{} predictions v{}".format(predictionType, self.version))
    predFileInfoText += "* {:<48s} *\n".format("")
    for infoText in self.infoTexts:
      predFileInfoText += "* {:<48s} *\n".format(infoText)

    return predFileInfoText[:-1]

  def getPredFileReferences(self):

    refText = ""
    for refDoi in self.references:
      refText += "* {:<48s} *\n".format(refDoi)

    return refText[:-1]

  def getPredFileHeader(self):

    predFileHeader = """
****************************************************
{}
*                                                  *
* for {:<45s}*
* on {:<46s}*
*                                                  *
* If you use these data please cite:               *
{}
****************************************************
"""

    return predFileHeader

  #
  # Multiple sequence alignment reading code
  #

  def readAlignments(self, fileName, fileType=None, resetAlignRefSeqID=False, gapCode='-'):

    # This is the gap code used in the alignment file
    self.gapCode = gapCode

    # Reset if required
    if resetAlignRefSeqID:
      self.alignRefSeqID = None

    # Read the file
    fin = open(fileName)
    lines = fin.readlines()
    fin.close()

    # Check which type of alignment file
    if not fileType:
      numLines = len(lines)
      if lines[0].count("CLUSTAL"):
        fileType = "CLUSTAL"
      elif lines[0].startswith("# STOCKHOLM"):
        fileType = 'STOCKHOLM'
      else:
        fastaCount = balibaseCount = psiCount = blastCount = emptyLine = fastaGaps = 0
        for line in lines:
          if line.startswith(">"):
            fastaCount += 1
          elif line.startswith("//"):
            balibaseCount += 1
          elif len(line.split()) == 2:
            psiCount += 1
          elif line.startswith("Query") or line.startswith("Sbjct"):
            blastCount += 1
          elif not line.strip():
            emptyLine += 1
          elif fastaCount and line.count(gapCode):
            fastaGaps += 1

        if fastaCount > balibaseCount and fastaCount > blastCount:
          if fileName.endswith('a3m'):
            fileType = 'A3M'
          elif fastaGaps:
            fileType = 'FASTA'
        elif blastCount:
          fileType = 'Blast'
        elif balibaseCount:
          fileType = 'BaliBase'
        elif psiCount == numLines - emptyLine:
          fileType = 'PSI'
        else:
          # Check for PHYLIP format
          cols = lines[0].split()
          if len(cols) == 2 and cols[0].isdigit() and cols[1].isdigit():
            fileType = 'PHYLIP'

    assert fileType, "Alignment file not recognised"

    print("Reading {} alignment...".format(fileType))

    self.alignRefSeqID = None
    if fileType == 'CLUSTAL':
      seqAlignments = self.readAlignmentsClustal(lines)
    elif fileType == 'FASTA':
      seqAlignments = self.readAlignmentsFasta(lines)
    elif fileType == 'BaliBase':
      seqAlignments = self.readAlignmentsBalibase(lines)
    elif fileType == 'PSI':
      seqAlignments = self.readAlignmentsPSI(lines)
    elif fileType == 'A3M':
      seqAlignments = self.readAlignmentsA3M(lines)
    elif fileType == 'Blast':
      seqAlignments = self.readAlignmentsBlast(lines)
    elif fileType == 'PHYLIP':
      seqAlignments = self.readAlignmentsPHYLIP(lines)
    elif fileType == 'STOCKHOLM':
      seqAlignments = self.readAlignmentsStockholm(lines)

    return seqAlignments

  def readAlignmentsFasta(self, lines):

    """
    FASTA file alignment
    """

    startReading = True
    seqAlignments = {}

    for line in lines:

      cols = line.split()

      if cols:

        if cols[0].startswith('>'):
          seqId = self.getSeqIdKey(cols[0][1:])
          if not self.alignRefSeqID:
            self.alignRefSeqID = seqId
        else:
          if seqId not in seqAlignments.keys():
            seqAlignments[seqId] = cols[0].upper()
          else:
            # Multiline FASTA
            seqAlignments[seqId] += cols[0].upper()

      else:
        self.setEmptyLineVars()

    return seqAlignments

  def readAlignmentsA3M(self, lines):

    """
    A3M file alignment, need some magic here to align things decently
    """

    startReading = True
    sequences = {}

    for line in lines:

      cols = line.split()

      if cols:

        if cols[0].startswith('>'):
          seqId = self.getSeqIdKey(cols[0][1:])
          if not self.alignRefSeqID:
            self.alignRefSeqID = seqId
        else:
          if seqId not in sequences.keys():
            sequences[seqId] = cols[0]
          else:
            # Multiline FASTA
            sequences[seqId] += cols[0]

      else:
        self.setEmptyLineVars()

    # Now reset sequence indexing. Not trivial, have to track numbering per sequence
    seqIndexes = {}
    seqAlignments = {}

    seqIds = [*sequences.keys()]
    numSeqs = len(seqIds)
    for seqId in seqIds:
      seqIndexes[seqId] = 0
      seqAlignments[seqId] = ""

    # Base on first sequence, is reference!
    while seqIndexes[self.alignRefSeqID] < len(sequences[self.alignRefSeqID]):
      currentColChars = "".join([sequences[seqId][seqIndexes[seqId]] for seqId in seqIds])

      # Clean, no inserts
      if currentColChars == currentColChars.upper():
        for i in range(numSeqs):
          seqAlignments[seqIds[i]] += currentColChars[i]
          seqIndexes[seqIds[i]] += 1
      else:
        for i in range(numSeqs):
          if currentColChars[i].islower():
            seqAlignments[seqIds[i]] += currentColChars[i].upper()
            seqIndexes[seqIds[i]] += 1
          else:
            seqAlignments[seqIds[i]] += self.gapCode

    return seqAlignments

  def readAlignmentsBlast(self, lines):

    print("Warning: BLAST format, not the full sequences in the alignment!")

    sequences = {}
    sequencesQuery = {}

    refQuerySeqId = ""

    maxQueryLen = 0

    for line in lines:

      cols = line.split()

      if cols:

        if cols[0].startswith('>'):
          seqId = self.getSeqIdKey(cols[0][1:])
          if not refQuerySeqId:
            refQuerySeqId = seqId
        elif cols[0].startswith("Query") and cols[0][-1] != '=':
          if seqId not in sequences.keys():
            seqStart = int(cols[1])

            sequences[seqId] = self.gapCode * (seqStart - 1) + ""
            sequencesQuery[seqId] = "X" * (seqStart - 1) + cols[2]
          else:
            sequencesQuery[seqId] += cols[2]
        elif cols[0].startswith("Sbjct"):
          sequences[seqId] += cols[2]

      else:
        self.setEmptyLineVars()

    # Also track query sequence
    self.alignRefSeqID = 'query'
    sequences[self.alignRefSeqID] = sequencesQuery[refQuerySeqId].replace(self.gapCode, "")
    sequencesQuery[self.alignRefSeqID] = sequences['query']
    querySeqLen = len(sequencesQuery[self.alignRefSeqID])

    seqLens = {}

    # Pad all other query seqs, might be bits missing at the end
    for seqId in sequencesQuery.keys():
      seqLens[seqId] = len(sequences[seqId])

    maxSeqLen = max(seqLens.values())

    for seqId in sequencesQuery.keys():
      if len(sequencesQuery[seqId]) < maxSeqLen:
        seqLenDiff = (maxSeqLen - len(sequencesQuery[seqId])) + 1
        sequencesQuery[seqId] += "X" * seqLenDiff
        sequences[seqId] += self.gapCode * seqLenDiff

    # Now reset sequence indexing. Not trivial, have to track numbering per sequence
    seqIndexes = {}

    seqAlignments = {}

    seqIds = [*sequences.keys()]
    numSeqs = len(seqIds)
    for seqId in seqIds:
      seqIndexes[seqId] = 0
      seqAlignments[seqId] = ""

    # Base on query sequence, is reference!
    while max(seqIndexes.values()) <= maxSeqLen:

      # Inserts in query means that all alignments that DO NOT have an insert here need one
      try:
        currentColCharsQuery = "".join([sequencesQuery[seqId][seqIndexes[seqId]] for seqId in seqIds])
      except:
        for seqId in seqIds:
          # if seqIndexes[seqId] == len(sequencesQuery[seqId]):
          print("{:<40s} {} {} {}".format(seqId, seqIndexes[seqId], len(sequencesQuery[seqId]), len(sequences[seqId])))
        raise

      if currentColCharsQuery.count(self.gapCode):
        for i in range(numSeqs):
          seqId = seqIds[i]
          if currentColCharsQuery[i] == self.gapCode:
            try:
              seqAlignments[seqId] += sequences[seqId][seqIndexes[seqId]]
              seqIndexes[seqId] += 1

            except:
              print(
                "{:<40s} {} {} {}".format(seqId, seqIndexes[seqId], len(sequencesQuery[seqId]), len(sequences[seqId])))
              raise

          else:
            seqAlignments[seqId] += self.gapCode

      else:
        for i in range(numSeqs):
          seqId = seqIds[i]
          try:
            seqAlignments[seqId] += sequences[seqId][seqIndexes[seqId]]
            seqIndexes[seqId] += 1
          except:
            print(
              "{:<40s} {} {} {}".format(seqId, seqIndexes[seqId], len(sequencesQuery[seqId]), len(sequences[seqId])))
            raise

    return seqAlignments

  def readAlignmentsBalibase(self, lines):

    """
    BaliBase alignment
    """

    startReading = False
    seqAlignments = {}

    for line in lines:

      if line.startswith("//"):
        startReading = True
        continue

      if startReading:
        cols = line.split()

        if cols:
          seqId = self.getSeqIdKey(cols[0])
          alignment = ''.join(cols[1:])

          if not self.alignRefSeqID:
            self.alignRefSeqID = seqId

          if seqId not in seqAlignments.keys():
            seqAlignments[seqId] = ""

          seqAlignments[seqId] += alignment

        else:
          self.setEmptyLineVars()

    return seqAlignments

  def readAlignmentsClustal(self, lines, uniqueSeqs=False):

    """
    CLUSTAL files

    If uniqueSeqs is True, will add extra suffix to overlapping identifiers occuring more than once, so they end up separately
    """

    startReading = False
    seqAlignments = {}

    for line in lines:

      if line.startswith("CLUSTAL"):
        startReading = True
        continue

      if startReading:
        cols = line.split()

        if cols:
          if len(cols) in (2, 3):

            # Ignore lines with annotation information
            if cols[0][0].count('*') or cols[0][0].count(":") or cols[0][0].count(".") or cols[0].isdigit():
              continue

            seqId = self.getSeqIdKey(cols[0])

            if uniqueSeqs and seqId in seqAlignments.keys():
              for i in range(99):
                newSeqId = "{}_{}".format(seqId, i)
                if newSeqId not in seqAlignments.keys():
                  seqId = newSeqId
                  break

            if not self.validSeqId(seqId):
              continue

            if not self.alignRefSeqID:
              self.alignRefSeqID = seqId

            alignment = cols[1]

            if seqId not in seqAlignments.keys():
              seqAlignments[seqId] = ""

            seqAlignments[seqId] += alignment

        else:
          self.setEmptyLineVars()

    return seqAlignments

  def readAlignmentsPSI(self, lines):

    """
    PSI file alignment
    """

    seqAlignments = {}

    for i in range(len(lines)):

      line = lines[i]

      if not line.strip():
        continue

      try:
        (seqId, sequence) = line.split()

        seqId = self.getSeqIdKey(seqId)

        if not self.alignRefSeqID:
          self.alignRefSeqID = seqId

        # For some reason additional - at beginning of sequence for target protein
        if i == 0 and sequence[0] == self.gapCode:
          sequence = sequence[1:]

        if seqId not in seqAlignments.keys():
          seqAlignments[seqId] = ""

        seqAlignments[seqId] += sequence
      except ValueError:
        continue

    return seqAlignments

  def readAlignmentsPHYLIP(self, lines):

    """
    PSI file alignment
    """

    seqAlignments = {}

    seqCounterToId = {}
    seqCounter = 0
    seqStartIndex = 11

    fileInfo = lines[0].split()
    (totalSeqs, seqAlignLen) = [int(count) for count in fileInfo]

    for i in range(1, len(lines)):

      line = lines[i].strip()

      if not line:
        seqCounter = 0
        seqStartIndex = 0

      else:
        seqStartIndex = line.rfind(" ") + 1
        sequence = line[seqStartIndex:] # .replace(" ", "")

        if i <= totalSeqs:
          seqId = self.getSeqIdKey(line[:seqStartIndex].strip())
          seqCounterToId[seqCounter] = seqId
          if not self.alignRefSeqID:
            self.alignRefSeqID = seqId
          seqAlignments[seqId] = ""

        else:
          seqId = seqCounterToId[seqCounter]

        seqAlignments[seqId] += sequence

        seqCounter += 1

    assert totalSeqs == len(seqAlignments), "Mismatch in number of sequences detected and reported"
    assert seqAlignLen == len(seqAlignments[seqId]), "Mismatch in length of alignment detected and reported"

    return seqAlignments

  def readAlignmentsStockholm(self, lines):

    """
    STOCKHOLM alignment file
    """

    seqAlignments = {}

    for line in lines:

      cols = line.split()

      if cols:
        if not cols[0].startswith('#') and not cols[0].startswith("//"):
          seqId = self.getSeqIdKey(cols[0])
          if not self.alignRefSeqID:
            self.alignRefSeqID = seqId
          seqAlignments[seqId] = cols[1].upper()

    return seqAlignments

  def setEmptyLineVars(self):

    # Custom function for subclasses
    pass

  def validSeqId(self, seqId):

    # Checks whether this is a valid sequence ID, can be used for filtering in subclasses
    return True

  def getSeqIdKey(self, seqId):

    # Make a good sequence ID from what's found in the alignment file; used in subclass
    return seqId

  def writeFastaFromAlignment(self, alignFile, outFastaFile):

    #
    # Writes a FASTA file from an alignment
    #

    seqAlignments = self.readAlignments(alignFile)

    self.writeFastaFromSeqAlignmentDict(seqAlignments, outFastaFile)

  def writeFastaFromSeqAlignmentDict(self, seqAlignments, outFastaFile):

    #
    # Writes a FASTA file from an alignment
    #

    seqIds = [*seqAlignments.keys()]
    seqIds.sort()

    fout = open(outFastaFile, 'w')
    for seqId in seqIds:
      fout.write(">{}\n".format(seqId))
      fout.write("{}\n".format(seqAlignments[seqId].replace(self.gapCode, '')))
    fout.close()

  #
  # JSON dumps
  #

  def retrieve_seq_uniprot(uniprotID):

    baseUrl = "http://www.uniprot.org/uniprot/"
    currentUrl = baseUrl + uniprotID + ".fasta"
    response = r.post(currentUrl)
    cData = ''.join(response.text)

    Seq = StringIO(cData)
    pSeq = self.readFasta(None,fileString=Seq)
    sequence = [(uniprotID, ''.join([res for res in pSeq[0]]))]

    return(sequence)

  def round_floats(self, o):
    if isinstance(o, float): return round(o, 3)
    if isinstance(o, dict): return {k: self.round_floats(v) for k, v in o.items()}
    if isinstance(o, (list, tuple)): return [self.round_floats(x) for x in o]
    return o

  def getAllPredictionsJson(self, identifier, limitPrecision=True):

    """
    Creates JSON string with all available predictions with identifier as base name
    :param identifier: Base name for the output files
    :return: JSON with full information for all predictions done in the subclass
    """

    jsonData = {
      'creation_date': time.strftime("%Y-%m-%dT%H:%M:%S"),
      'id': identifier,
      'information': self.informationPerPredictor,
      'results': []
    }

    seqIds = list(self.allPredictions.keys())
    seqIds.sort()

    predictionTypes = list(self.allPredictions[seqIds[0]].keys())

    for seqId in seqIds:
      sequence = ''.join([resInfo[0] for resInfo in self.allPredictions[seqId][predictionTypes[0]]])
      seqInfo = {'proteinID': seqId, 'sequence': sequence}

      for predictionType in predictionTypes:
        seqInfo[predictionType] = [resInfo[1] for resInfo in self.allPredictions[seqId][predictionType]]
        #print(seqId,predictionType,type(seqInfo[predictionType][0]),seqInfo[predictionType])

      jsonData['results'].append(seqInfo)

    # Limit precision in output, 3 decimals is plenty for this. Doesn't seem to work on Python3.7 any more
    if limitPrecision:
      json.encoder.FLOAT_REPR = lambda o: format(o, '.3f')


    # return(json.dumps(jsonData, sort_keys=True, indent=4, separators=(',', ': ')))
    return json.dumps(self.round_floats(jsonData), sort_keys=True, indent=4, separators=(',', ': '))

  def getAllPredictionsJson_msa(self, results, limitPrecision=True):

      """
      Creates JSON string with all available predictions with identifier as base name
      :param identifier: Base name for the output files
      :return: JSON with full information for all predictions done in the subclass
      """

      jsonData = {
        'creation_date': time.strftime("%Y-%m-%dT%H:%M:%S"),
        'information': self.informationPerPredictor,
        'results': results
      }

      json_data = json.dumps(self.round_floats(jsonData), sort_keys=True, indent=4, separators=(',', ': '))
      return jsonData

  def json_to_fasta_variants(json_filename, output_dir=""):
    with open(json_filename) as json_file:
      data = json.load(json_file)
      # Generate fasta file with the WT & variants
      filename = os.path.join(output_dir, "{}.fasta".format(data["metadata"]["name"]))
      fasta_file = open(filename, "w")
      # Add the WT
      fasta_file.write(">WT\n")
      fasta_file.write(data["WT"] + "\n")
      # Add the variants
      for key in data["Variants"].keys():
        fasta_file.write(">{}\n".format(key))
        # Add all the mutations per variant
        mut_seq = list(data["WT"])
        adp_val = 0 # Value to adapt the numbering affected by insertions
        for mut in data["Variants"][key]:
          # Deletions
          if "del" in mut:
            pos = int(mut[1:-3]) - 1 + adp_val
            mut_seq[pos] = "-"

          # Insertions
          elif "ins" in mut:
            pos = int(mut.split("_")[0][1:]) + adp_val

            for aa in mut.split("ins")[1]:
              mut_seq.insert(pos + adp_val, aa)
              adp_val += 1

          else:
            # Identify the position to be mutated
            pos = int(mut[1:-1]) - 1 + adp_val
            # Mutate the residue
            mut_seq[pos] = mut[-1]

        # Remove deletion "-" sings if any
        mut_seq = [i for i in mut_seq if i != "-"]

        # Add the mutated sequences in the fasta file
        fasta_file.write("{}\n".format("".join(mut_seq)))

    fasta_file.close()
    # Run the MSA using all the BLAST matches found
    aligned_file = '{}_tcof.aln'.format(filename.split('.')[0])
    cmd = "t_coffee {} -output=fasta_aln -outfile={}".format(
      filename, aligned_file)

    os.system(cmd)
    return aligned_file

  #
  # NEF file handling
  #

  def readNefFile(self,fileName):

    """
    :param fileName: Input NEF file path
    :return: File object with info, if read, otherwise None.
    """

    from .bmrb import File as starReader

    # Note that this nmrStar reader is generic STAR, not specifically for NMR-STAR.
    # If this works, could also be an NMR-STAR file!
    origNefFile = starReader.File(verbosity=0, filename=fileName)

    #
    # Read NEF file (in STAR format)
    #

    if origNefFile.read():
      print("  Error reading NEF file ")
      return None
    else:
      if not origNefFile.datanodes:
        print("  No NEF data available.")
        return None

    return origNefFile

  def readNefFileSequenceShifts(self,fileName):

    """

    Note: this reader is limited, as it only handles the sequence and chemical shift information.
    It will need to be extended for other data types, if that should become necessary.

    :param fileName: Input NEF file path
    :return: dictionary with chainCode as key, then a list of sequence codes (pos 0) and a list of and shift information (pos 1)
    """

    origNefFile = self.readNefFile(fileName)

    sequenceInfo = {}

    if origNefFile:

      #
      # Go through the data in the file
      #

      seqCodeToIndex = {}
      seqSerialToIndex = {}

      currentSaveFrame = None

      for origSaveFrame in origNefFile.datanodes:

        # This is the title of the saveframe
        # print origSaveFrame.title

        for tagtable in origSaveFrame.tagtables:

          if tagtable.free:

            # This is for values directly associated with saveframe, only one value per tagname

            for tagIndex in range(len(tagtable.tagnames)):

              tagName = tagtable.tagnames[tagIndex]
              tagValue = tagtable.tagvalues[tagIndex][0]  # Only one value, always!

              # print tagName, tagValue

              if tagName.endswith('sf_category'):
                currentSaveFrame = tagValue
                # print currentSaveFrame

          else:

            # This is a loop with multiple rows of values for the tags
            # Have to loop over the value index to get rows out

            # print("Table with tags {}".format(",".join(tagtable.tagnames)))
            tags = [tagName.split('.')[-1] for tagName in tagtable.tagnames]

            # print tags

            numTagIndexes = len(tagtable.tagnames)

            # Reading the sequence, comes first
            if currentSaveFrame == 'nef_molecular_system':
              # Using first tag index (0) to get number of data value rows in loop
              seqListIndex = 0
              for valueIndex in range(len(tagtable.tagvalues[0])):
                dataRow = {}
                for tagIndex in range(numTagIndexes):
                  tagValue = tagtable.tagvalues[tagIndex][valueIndex]
                  dataRow[tags[tagIndex]] = tagValue

                chainCode = dataRow['chain_code']
                if chainCode not in sequenceInfo.keys():
                  sequenceInfo[chainCode] = []
                  seqListIndex = 0

                sequenceInfo[chainCode].append((dataRow, []))  # Second list will hold shifts

                seqCodeToIndex[(chainCode,dataRow['sequence_code'])] = seqListIndex
                seqSerialToIndex[(chainCode,dataRow['index'])] = seqListIndex
                seqListIndex += 1

            elif currentSaveFrame == 'nef_chemical_shift_list':
              # print(seqCodeToIndex)

              # Reading shift info and linking to full sequence information
              # Put in a hack below to deal with badly formatted NEF files

              seqCodesValid = True
              for valueIndex in range(len(tagtable.tagvalues[0])):
                dataRow = {}
                for tagIndex in range(numTagIndexes):
                  tagValue = tagtable.tagvalues[tagIndex][valueIndex]
                  dataRow[tags[tagIndex]] = tagValue

                # print(dataRow)
                chainCode = dataRow['chain_code']
                seqCode = dataRow['sequence_code']

                if (chainCode,seqCode) not in seqCodeToIndex.keys():
                  seqCodesValid = False

              # Now read full file
              for valueIndex in range(len(tagtable.tagvalues[0])):
                dataRow = {}
                for tagIndex in range(numTagIndexes):
                  tagValue = tagtable.tagvalues[tagIndex][valueIndex]
                  dataRow[tags[tagIndex]] = tagValue

                # print(dataRow)
                chainCode = dataRow['chain_code']
                seqCode = dataRow['sequence_code']
                if seqCodesValid:
                  seqListIndex = seqCodeToIndex[(chainCode,seqCode)]
                else:
                  # Is really the index here, badly formatted NEF file
                  seqListIndex = seqSerialToIndex[(chainCode,seqCode)]

                # Adding relevant shift info to file, removing redundant stuff first
                del (dataRow['chain_code'])
                del (dataRow['residue_name'])
                del (dataRow['sequence_code'])
                sequenceInfo[chainCode][seqListIndex][-1].append(dataRow)

    return sequenceInfo

  #
  # NMR-STAR file handling
  #

  def readNmrStarProject(self,fileName):

    """
    :param filenNme:  Input NMR-STAR file path
    :return: file content connected to nmrStarFile object
    """

    from b2bTools.general.ccpn.format.nmrStar.projectIO import NmrStarProjectFile

    nmrStarFile = NmrStarProjectFile(fileName)
    readStatus = nmrStarFile.read(verbose=1)

    if not readStatus or not nmrStarFile.sequenceFiles:
      # Not a valid file, reset to None
      nmrStarFile = None

    return nmrStarFile

  def readNmrStarSequenceShifts(self,fileName, original_numbering=True):

    """
    :param fileName: Input NMR-STAR file path
    :param original_numbering: If set to True (Boolean), will retain original sequence code numbering
    :return: dictionary with sequence and shift information
    """

    # Legacy hack for web server side code
    if type(original_numbering) == type(""):
      original_numbering = eval(original_numbering)

    nmrStarFile = self.readNmrStarProject(fileName)

    sequenceInfo = {}

    if nmrStarFile and nmrStarFile.sequenceFiles:

      seqCodeToIndex = {}
      seqSerialToIndex = {}

      # Get protein sequence data out
      chainCodes = 'ABCDEFGHIJKLM'
      chainCodeIndex = 0

      molCodeToChainCode = {}
      seqOffset = 0
      for seqFile in nmrStarFile.sequenceFiles:
        for seq in seqFile.sequences:
          if seq.molType != 'polymer':
            continue
          # print(dir(seq))
          if seq.molName not in molCodeToChainCode.keys():
            currentChainCode = chainCodes[chainCodeIndex]
            molCodeToChainCode[seq.molName] = currentChainCode
            chainCodeIndex += 1

          chainCode = molCodeToChainCode[seq.molName]
          sequenceInfo[chainCode] = []

          valueIndex = 0
          for sequenceEl in seq.elements:
            if original_numbering and hasattr(sequenceEl,'authorSeqCode') and sequenceEl.authorSeqCode != None:
              seqNumber = sequenceEl.authorSeqCode
              if sequenceEl.seqCode == 1:
                seqOffset = int(sequenceEl.authorSeqCode) - 1
            else:
              seqNumber = sequenceEl.seqCode

            # Converting to NEF-like structure here
            dataRow = {'index': sequenceEl.seqCode, 'chain_code': currentChainCode, 'sequence_code': seqNumber, 'residue_name': sequenceEl.code3Letter}

            sequenceInfo[chainCode].append((dataRow, []))

            seqCodeToIndex[dataRow['sequence_code']] = valueIndex
            seqSerialToIndex[dataRow['index']] = valueIndex

            valueIndex += 1

      # Get chemical shift data, bunching all together here in assumption only one saveframe for this
      for chemShiftFile in nmrStarFile.chemShiftFiles:
        for chemShift in chemShiftFile.chemShifts:
          if chemShift.molCode in molCodeToChainCode.keys():
            chainCode = molCodeToChainCode[chemShift.molCode]
          else:
            chainCode = 'A'

          if original_numbering:
            seqNumber = chemShift.seqCode + seqOffset
          else:
            seqNumber = chemShift.seqCode

          dataRow = {'atom_name': chemShift.atomName, 'value': chemShift.value, 'value_uncertainty': chemShift.valueError}

          # print(dataRow)
          seqCode = seqNumber
          seqIndex = seqCodeToIndex[seqNumber]
          sequenceInfo[chainCode][seqIndex][-1].append(dataRow)

    return sequenceInfo

  def convertNmrStarToNef(self,fileName):

    # Legacy code from web server, works only for shifts and sequence info!!

    # Get protein sequence data out
    chainCodes = 'ABCDEFGHIJKLM'
    chainCodeIndex = 0


    sequenceData = StringIO()


    molCodeToChainCode = {}
    seqOffset = 0
    for seqFile in nmrStarFile.sequenceFiles:
      for seq in seqFile.sequences:
        if seq.molType != 'polymer':
          continue
        # print(dir(seq))
        if seq.molName not in molCodeToChainCode.keys():
          currentChainCode = chainCodes[chainCodeIndex]
          molCodeToChainCode[seq.molName] = currentChainCode
          chainCodeIndex += 1
        for sequenceEl in seq.elements:
          if original_numbering == 'True' and sequenceEl.authorSeqCode != None:
            seqNumber = sequenceEl.authorSeqCode
            if sequenceEl.seqCode == 1:
              seqOffset = int(sequenceEl.authorSeqCode) - 1
          else:
            seqNumber = sequenceEl.seqCode

            dataRow = {'index': sequenceEl.seqCode, 'chain_code': currentChainCode, 'sequence_code': seqNumber, 'residue_name': sequenceEl.code3Letter}

          sequenceData.write(
            "                {:4d}  {:s} {:4d} {:3s}  .     .\n".format(sequenceEl.seqCode, currentChainCode,
                                                                        int(seqNumber), sequenceEl.code3Letter))

    # Get chemical shift data, bunching all together here in assumption only one saveframe for this


    shiftData = StringIO()


    for chemShiftFile in nmrStarFile.chemShiftFiles:
      for chemShift in chemShiftFile.chemShifts:
        if chemShift.molCode in molCodeToChainCode.keys():
          chainCode = molCodeToChainCode[chemShift.molCode]
        else:
          chainCode = 'A'

        if original_numbering == 'True':
          seqNumber = chemShift.seqCode + seqOffset
        else:
          seqNumber = chemShift.seqCode

        shiftData.write(
          "                {:s}   {:4d}    {:3s}   {:4s}  {:7.2f}    {:5.2f}\n".format(chainCode, int(seqNumber),
                                                                                       chemShift.resLabel,
                                                                                       chemShift.atomName,
                                                                                       chemShift.value,
                                                                                       chemShift.valueError))

  def json_preds_to_csv_singleseq(self, json_data, output_dir="results"):
      # Function to save the predictions in the json file to a txt-like file
      filename = json_data["id"].split("/")[-1].split(".")[0]
      output_file = os.path.join(output_dir, '{}_b2btools_preds.txt'.format(filename))

      with open(output_file, 'w+') as predfile:		# outfile here
        writer = csv.writer(predfile, delimiter=',')

        headerWritten = False
        for preds in json_data["results"]:  # iterate the predictions and get information to write to csv
          seqid = preds["proteinID"]

          predNames = ('sequence', 'backbone', 'sidechain', 'sheet', 'helix', 'coil', 'earlyFolding', 'disoMine', 'agmata')
          predNamesFound = []
          predsFound = []

          for predName in predNames:
            if predName in preds.keys():
              predNamesFound.append(predName)

              if predName in ('sequence'):
                predsFound.append([elem for elem in preds[predName]])
              else:
                predsFound.append([getCsvString(elem) for elem in preds[predName]])

          # WV to fix!
          # Fix sequence to alignment, if this is a msa-connected prediction
          #if tool == 'msatoolsproteins':
          #  seqNumberCode = 'AlignCode'
          #  for i in range(len(predsFound[1])):
              # If None, is a gap, so insert None in sequence as well
          #    if predsFound[1][i] == 'None':
          #      predsFound[0].insert(i, '-')
          #else:
          seqNumberCode = 'SeqCode'

          seqLen = len(predsFound[0])  # Sequence length

          # Insert sequence numbers
          predNamesFound.insert(0, seqNumberCode)
          predsFound.insert(0, range(1, seqLen + 1))

          # Insert protein info
          predNamesFound.insert(0, "Accession")
          predsFound.insert(0, [seqid] * seqLen)

          if not headerWritten:
            writer.writerow(predNamesFound)
            headerWritten = True

          allpredsperseq = zip(*predsFound)
          for row in allpredsperseq:
            writer.writerow(row)

  def json_preds_to_csv_msa(self, json_data, id, output_dir = 'results'):
        # Function to save the predictions in the json file to a txt-like file
        output_file = os.path.join(output_dir, '{}_b2btools_preds.txt'.format(id))

        with open(output_file, 'w+') as predfile:		# outfile here
          preds = json_data["results"]["results"]
          #preds = response_dict['statistic']

          writer = csv.writer(predfile, delimiter=',')

          predCodes = list(preds.keys())
          predCodes.sort()

          predNames = ('backbone', 'sidechain', 'sheet', 'helix', 'coil', 'earlyFolding', 'disoMine', 'agmata')
          statsNames = ("median", "topOutlier", "firstQuartile", "thirdQuartile", "bottomOutlier")

          predNamesFound = []
          predsFound = []

          for predName in predNames:
            if predName in preds.keys():
              for statsName in statsNames:
                if statsName in preds[predName].keys():
                  predNamesFound.append("{}_{}".format(predName, statsName))
                  predsFound.append([getCsvString(elem) for elem in preds[predName][statsName]])

          alignLen = len(predsFound[0])  # Alignment length

          # Insert alignment numbers
          predNamesFound.insert(0, 'AlignCode')
          predsFound.insert(0, range(1, alignLen + 1))

          writer.writerow(predNamesFound)

          allpredsperseq = zip(*predsFound)
          for row in allpredsperseq:
            writer.writerow(row)



# -*- coding: utf-8 -*-
import os
import tempfile
import subprocess, sys

cwd = os.path.dirname(os.path.abspath(__file__))

DATADIR = cwd + '/psipred/data/'
# TEMPDIR = tempfile.mkdtemp() # cwd + '/psipred/tmp/'

if sys.platform == 'linux':
    SEQ2MTX = os.path.join(cwd, 'psipred/bin/linux/seq2mtx')
    PSIPRED = os.path.join(cwd, 'psipred/bin/linux/psipred')
    PSIPASS2 = os.path.join(cwd, 'psipred/bin/linux/psipass2')
elif sys.platform == 'darwin': # MacOS hosts
    SEQ2MTX = os.path.join(cwd, 'psipred/bin/osx/seq2mtx')
    PSIPRED = os.path.join(cwd, 'psipred/bin/osx/psipred')
    PSIPASS2 = os.path.join(cwd, 'psipred/bin/osx/psipass2')

for exec_path in [SEQ2MTX, PSIPRED, PSIPASS2]:
    if os.path.exists(exec_path):
        os.chmod(exec_path, int('777', 8))

def run(input, tmpfolder = None):
    if tmpfolder == None:
        tmpfolder = tempfile.mkdtemp()

    fasta = os.path.basename(input)
    name = fasta.split('.')[-2]

    mtx_filepath = os.path.join(tmpfolder, "{}.mtx".format(name))
    ss_filepath = os.path.join(tmpfolder, "{}.ss".format(name))
    ss2_filepath = os.path.join(tmpfolder, "{}.ss2".format(name))
    horiz_filepath = os.path.join(tmpfolder, "{}.horiz".format(name))

    # print('Psipred Running... processing: ' + name)

    with open(mtx_filepath, "w") as mtx_file:
        subprocess.run([SEQ2MTX, input], stdout=mtx_file)

    with open(ss_filepath, "w") as ss_file:
        subprocess.run( [PSIPRED, mtx_filepath, DATADIR + 'weights.dat', DATADIR + 'weights.dat2', DATADIR + 'weights.dat3'], stdout=ss_file)

    with open(horiz_filepath, "w") as horiz_file:
        subprocess.run( [PSIPASS2, DATADIR + 'weights_p2.dat', '1', '1.0', '1.0', ss2_filepath, ss_filepath], stdout=horiz_file)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  standalone.py
#
#  Copyright 2017 Gabriele Orlando <orlando.gabriele89@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

from b2bTools.singleSeq.DisoMine.vector_builder.vettore_gen import build_vector
from b2bTools.singleSeq.DisoMine.torch_NN_gru_80AUC_prova import nn_pytorch
import numpy as np
import sys
import pathlib
import string, random
import tempfile
import os

def standalone(seqList, dmPredictions, verbose=0):
    """
	seqList is now a list of tuples defined by (seqId, seq) which is faster.
	dmPredictions is the existing dynaMine/EFoldMine preds (done earlier).
	verbose defines the level of feedback given to the user printing messages on the standard output.
	"""

    # These are the features used in the final version
    features = 'dyna_back,psipred,ef,dyna_side'

    # Directory of this script
    scriptDir = str(pathlib.Path(__file__).parent.absolute())
    tempDir = tempfile.mkdtemp()
    mapfilepath = os.path.join(tempDir, 'map_random_name_with_seqID.tsv')
    print("Map file for random names and sequence ids: {}".format(mapfilepath))

    with open(mapfilepath, 'w') as f:
        f.write('random_name\tseq_id\tdisoMine_status\tDetails\n')

    # This is the window size used
    window = 4

    def check_sequences(seqList):
        """
		This method checks if all the provided sequences meet the minimum requirements:
		all the characters of the sequences are valid and the length of each sequence is greater than 20.

		If all the sequences meet the checks, a True value is returned, otherwise a error object.
		"""
        for (seqId, seq) in seqList:
            if not seq.isalpha():
                return {'error': 'invalid char in sequence {}'.format(seqId)}

            if len(seq) < 20:
                return {'error': 'sequence {} too short, minimum length is 20 amino acids'.format(seqId)}

        return True

    def load_model():
        sys.path.append(scriptDir)

        mod = nn_pytorch(cuda=False)
        # mod.load_model('{}/gru80_final.mtorch'.format(scriptDir))
        # mod.load_model('{}/conversion_attempt.mtorch'.format(scriptDir))
        mod.load_model('{}/disomine_converted.mtorch'.format(scriptDir))

        scaler = None

        return mod, scaler

    def format_output(disorder, seqList, dmPredictions):
        results = {}

        for (protID, sequence) in seqList:
            results[protID] = []

            for sequenceIndex in range(len(sequence)):
                result_tuple = (sequence[sequenceIndex], np.float64(disorder[protID][sequenceIndex]))
                results[protID].append(result_tuple)

            dmPredictions[protID]['disoMine'] = results[protID]

        return results

    def predict(seqList, dmPredictions, model, crunch=100):

        check = check_sequences(seqList)

        if not check:
            return check

        results_dict = {}
        # dyna = {}
        # side = {}
        # ef = {}
        sequence_counter = 0
        numSeqs = len(seqList)

        # Doing 100 sequences at a time with disomine
        for sequence_index in range(0, numSeqs + 1, crunch):
            if sequence_index + crunch > numSeqs:
                end = numSeqs
            else:
                end = sequence_index + crunch

            vector_crunch = []

            if verbose >= 1:
                print('starting crunch', sequence_counter)

            targets = []
            for (target, seq) in seqList[sequence_index:end]:

                diso_works = False
                base_name = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(7))

                if verbose >= 2:
                    print('\tstarting target:', target)
                try:
                    disomine_results = build_vector(seq, dmPredictions[target], base_name=base_name, seqID=target,
                                                    TYPE=features, sw=window, tmpfold=tempDir)
                    with open(mapfilepath, 'a') as f:
                        f.write('{0}\t{1}\tSuccess\t-\n'.format(base_name, target))
                    diso_works = True

                except BaseException as err:
                    print("PSIPRED - An unexpected error occured: ", err)
                    with open(mapfilepath, 'a') as f:
                        f.write('{0}\t{1}\tFailed\t{2}\n'.format(base_name, target, err))
                    results_dict[target] = np.array([np.nan]*len(seq))

                if diso_works:
                    vector = np.array(disomine_results)
                    vector_crunch += [vector]
                    targets.append(target)


            sequence_counter += 1
            predictions = model.decision_function(vector_crunch)
            assert len(predictions) == len(vector_crunch)

            for target, prediction_results in zip(targets, predictions):
                # Set floor and ceiling of values not included in (0, 1)
                for value_index in range(len(prediction_results)):
                    if prediction_results[value_index] < 0:
                        prediction_results[value_index] = 0.0
                    elif prediction_results[value_index] > 1:
                        prediction_results[value_index] = 1.0

                results_dict[target] = prediction_results

        # Adding results to existing information as well here, but output is just disoMine
        results = format_output(results_dict, seqList, dmPredictions)

        return results

    # This is the main code for this def (standalone), calling other defs inside this one
    model, _scaler = load_model()
    results = predict(seqList, dmPredictions, model)

    return results


def main(args):
    # print standalone("example.fasta")
    a = leggifasta('input_files_examples/example_toy.fasta')
    a['extra_predictions'] = False
    print(standalone(a))


# from memory_profiler import memory_usage
# mem_usage = memory_usage(standalone,interval=0.01)
# print('Memory usage (in chunks of .1 seconds): %s' % mem_usage)
# print('Maximum memory usage: %s' % max(mem_usage))
# cProfile.run('standalone("example.fasta")')

if __name__ == '__main__':
    import sys

    sys.exit(main(sys.argv))




"""
======================COPYRIGHT/LICENSE START==========================

Io.py: Utility code for CCPN code generation framework

Copyright (C) 2005  (CCPN Project)

=======================================================================

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

A copy of this license can be found in ../../../license/LGPL.license

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


======================COPYRIGHT/LICENSE END============================

for further information, please contact :

- CCPN website (http://www.ccpn.ac.uk/)

- email: ccpn@bioc.cam.ac.uk

=======================================================================

If you are using this software for academic purposes, we suggest
quoting the following references:

===========================REFERENCE START=============================
R. Fogh, J. Ionides, E. Ulrich, W. Boucher, W. Vranken, J.P. Linge, M.
Habeck, W. Rieping, T.N. Bhat, J. Westbrook, K. Henrick, G. Gilliland,
H. Berman, J. Thornton, M. Nilges, J. Markley and E. Laue (2002). The
CCPN project: An interim report on a data model for the NMR community
(Progress report). Nature Struct. Biol. 9, 416-418.

Rasmus H. Fogh, Wayne Boucher, Wim F. Vranken, Anne
Pajon, Tim J. Stevens, T.N. Bhat, John Westbrook, John M.C. Ionides and
Ernest D. Laue (2005). A framework for scientific data modeling and automated
software development. Bioinformatics 21, 1678-1684.

===========================REFERENCE END===============================

"""
#
# Convenient I/O functions
#

import os
import shutil
from io import StringIO as StringIO
import sys
import traceback
import glob

from b2bTools.general.ccpn.universal import Constants

def printError(title, message, parent=None):
  """ Error printing - replaces showError in non-GUI contexts
  """
  print ("%s:\n  %s" % (title, message))

def printExceptionAndTrace(fp = sys.stdout):

  sf = StringIO.StringIO()
  exc = sys.exc_info()
  tb = exc[2]
  traceback.print_tb(tb, file=sf)
  ss = sf.getvalue()
  sf.close()
  del sf
  del tb
  fp.write('%s: %s\n%s' % (exc[0].__name__, exc[1], ss))

def getExceptionAndTrace():

  sf = StringIO.StringIO()
  printExceptionAndTrace(fp=sf)
  ss = sf.getvalue()
  sf.close()

  return ss

# Note: makeAbsolute is Boolean so the default should be
# False but ObjectDomain does not understand True/False
def normalisePath(path, makeAbsolute=None):
  """
  Normalises the path, e.g. removes redundant .. and slashes and
  makes sure path uses '/' rather than '\\' as can happen on Windows.
  """

  dirsep = Constants.dirsep
  assert dirsep == '/', "have dirsep = '%s', not '/'" % dirsep

  # note, cannot just use os.sep below because can have window file names cropping up on unix machines
  winsep = '\\'

  if os.sep == winsep:
    path = path.replace(dirsep, winsep)
    path = os.path.normpath(path)
    path = path.replace(winsep, dirsep)
  else:
    path = path.replace(winsep, dirsep)
    path = os.path.normpath(path)

  if makeAbsolute:
    if path and path[0] != dirsep:
      path = joinPath(os.getcwd(), path)

  return path

def unnormalisePath(path):
  """
  On Unix does nothing, on Windows replaces '/' with '\\'/
  """

  if (Constants.dirsep != os.sep):
    path = path.replace(Constants.dirsep, os.sep)

  return path

def joinPath(*args):
  """
  The same as os.path.join but normalises the result.
  """

  return normalisePath(os.path.join(*args))

def splitPath(path):
  """
  The same as os.path.split but with normalisation taken into account.
  """

  (head, tail) = os.path.split(unnormalisePath(path))
  head = normalisePath(head)

  return (head, tail)

def converseSplitPath(path):
  """
  Similar to splitPath but with head being the top directory and tail being the rest.
  """

  pair = splitPath(path)
  head = None
  tail = ''
  while pair[0] and pair[0] not in ('.', '/'):
    head = pair[0]
    tail = joinPath(pair[1], tail)
    pair = splitPath(head)

  if head is None:
    (head, tail) = pair

  return (head, tail)

def getTopDirectory():

  """
  Returns the 'top' directory of the whole code and data structure.
  """

  func = os.path.dirname

  return func(getPythonDirectory())

def pathIsInDirectory(path, directory):
  """Does path point to a file that is inside directory, possibley indirectly
  NB does not check whether either exists
  """
  ss = normalisePath(path)
  directory = normalisePath(directory)
  if ss.startswith(directory):
    lendir = len(directory)
    while len(ss) > lendir:
      ss,junk = os.path.split(ss)
    return (ss == directory)
  else:
    return False



def setTopDirectory(newTopDirectory):

  """
  Set a new 'top' directory of the whole code and data structure.
  This is so that you can swap between different implementations.
  WARNING: This function should only be used by experts.
  """

  # There is no perfect way at getting at keys in sys.modules which
  # need to be deleted, because there are some funny keys for which
  # sys.modules[key] is None (otherwise you could use the __file__
  # attribute of the module).  So look at the directories in
  # the old python directory and see which of those have an __init__.py
  # inside and assume that those are the modules we are trying to remove

  newTopDirectory = normalisePath(newTopDirectory)
  oldTopDirectory = getTopDirectory()
  oldPythonDirectory = getPythonDirectory()
  # Below is a bit dangerous; could have used os.path.basename(oldPythonDirectory) instead
  # if could guarantee that python directory is only one directory below top directory
  n = len(oldTopDirectory)
  newPythonDirectory = newTopDirectory + oldPythonDirectory[n:]

  moduleNames = []
  files = os.listdir(oldPythonDirectory)
  for file in files:
    fullfile = joinPath(oldPythonDirectory, file)
    if os.path.isdir(fullfile):
      if '__init__.py' in os.listdir(fullfile):
        moduleNames.append(file)

  # Below does not necessarily find all paths we want because symbolic
  # links mean that sometimes the wrong names are stored in sys.paths

  paths = []
  for path in sys.path:
    path = normalisePath(path)
    if path == oldTopDirectory or path.startswith(oldTopDirectory + Constants.dirsep):
      paths.append(path)

  sys_modules = sys.modules
  sys_path = sys.path

  keys = sys_modules.keys()
  for key in keys:
    for name in moduleNames:
      if key == name or key.startswith(name + '.'):
        del sys_modules[key]
        break

  for path in paths:
    sys_path.remove(path)

  # Now that have deleted what we can from sys.modules and sys.paths, put in
  # new information in sys.paths about newTopDirectory

  sys_path.insert(0, newPythonDirectory)

def getPythonDirectory():

  """
  Returns the 'top' python directory, the one on the python path.
  """

  func = os.path.dirname

  return func(func(func(os.path.abspath(Constants.__file__))))

def removePath(path):
  """Removes path whether file or directory, taking into account whether symbolic link.
  """

  if not os.path.exists(path):
    return

  if os.path.isfile(path) or os.path.islink(path):
    os.remove(path)
  else:
    shutil.rmtree(path)

def parentDirectory(directory):
  """parent directory
  """

  (parent_dir, tail_dir) = splitPath(directory)

  if (not tail_dir): # means have trailing slash, so do it without that
    parent_dir = splitPath(parent_dir)[0]

  return parent_dir


def writePythonDir(dirname):
  """ make python directory dirname (unless it exists already)
  and write blank __init__.py into it, if none exists.
  """

  # make directory
  if not os.path.exists(dirname):
    os.mkdir(dirname)

  # write __init__ file
  initfile = joinPath(dirname,'__init__.py')
  if not os.path.exists(initfile):
    open(initfile,'w').write("pass\n")

def importModule(name):
  """ Imports module in same way as __import__ but for modules of
  form x.y returns x.y not x (as __import__ does)
  """

  m = __import__(name)

  comps = name.split('.')
  for comp in comps[1:]:
    m = getattr(m, comp)

  return m

def returnRealPath(fileName):

  if os.path.islink(fileName):
    fileName = os.path.realpath(fileName)

  return fileName

def convertStringToFileName(fileNameString):

  from b2bTools.general.ccpn.general.Constants import validCcpnFileNameChars
  from b2bTools.general.ccpn.general.Constants import defaultFileNameChar

  return doConvertStringToFileName(fileNameString, validCcpnFileNameChars,
                                   defaultFileNameChar)


def doConvertStringToFileName(fileString, validChars, defaultChar):

  ll = [x for x in fileString]
  for ii,char in enumerate(ll):
    if char not in validChars:
      ll[ii] = defaultChar
  #
  return ''.join(ll)

def commonSuperDirectory(*fileNames):
  """ Find lowest directory that contains all files in list
  NB does not normalise file names.

  Input: a list of file names
  Output: lowest directory that contains all files. Does *not* end with a file
  """
  return os.path.dirname(os.path.commonprefix(fileNames))

def suggestFileLocations(fileNames, startDir=None):
  """ From a list of files, return a common superdirectory and a list of
  relative file names. If any of the files do not exist, search for an
  alternative superdirectory that does contain the set of relative file names.
  Searches in either a auperdirectory of the starting/current directory,
  or in a direct subdirectory.

  Input: list of file names
  Output: Superdirectory, list of relative file names.
          If no suitable location is found, superdirectory is returned as None
  """

  if not fileNames:
    return None, []

  # set up startDir
  if startDir is None:
    startDir = os.getcwd()
  startDir = normalisePath(startDir, makeAbsolute=True)

  # find common baseDir and paths
  files = [normalisePath(fp, makeAbsolute=True) for fp in fileNames]
  baseDir = commonSuperDirectory(*files)
  prefix = os.path.join(baseDir, '')
  lenPrefix = len(prefix)
  paths = [fp[lenPrefix:] for fp in files]

  if [fp for fp in files if not os.path.exists(fp)]:
    # some file not found.

    # look in superdirectories
    tail = 'junk'
    baseDir = startDir
    while tail:
      for path in paths:
        fp = os.path.join(baseDir, path)
        if not os.path.exists(fp):
          # a file is not found. try new baseDir
          break
      else:
        # this one is OK - stop searching
        break
      #
      baseDir, tail = os.path.split(baseDir)

    else:
      # No success - try in a subdirectory (one level) of startDir
      matches = glob.glob(os.path.join(startDir, '*', paths[0]))
      for aMatch in matches:
        baseDir = normalisePath(aMatch[:-len(paths[0])])
        for path in paths:
          fp = os.path.join(baseDir, path)
          if not os.path.exists(fp):
            # a file is not found. try new baseDir
            break
        else:
          # this one is OK - stop searching
          break
      else:
        # we give up
        baseDir = None
  #
  return (baseDir, paths)


class RedirectToString:

    def __init__(self):
      self.string = ""

    def write(self, s):
      self.string += s

    def close(self):
      pass




"""
======================COPYRIGHT/LICENSE START==========================

Util.py: Utility code for CCPN code generation framework

Copyright (C) 2005  (CCPN Project)

=======================================================================

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.
 
A copy of this license can be found in ../../../license/LGPL.license
 
This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.
 
You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA


======================COPYRIGHT/LICENSE END============================

for further information, please contact :

- CCPN website (http://www.ccpn.ac.uk/)

- email: ccpn@bioc.cam.ac.uk

=======================================================================

If you are using this software for academic purposes, we suggest
quoting the following references:

===========================REFERENCE START=============================
R. Fogh, J. Ionides, E. Ulrich, W. Boucher, W. Vranken, J.P. Linge, M.
Habeck, W. Rieping, T.N. Bhat, J. Westbrook, K. Henrick, G. Gilliland,
H. Berman, J. Thornton, M. Nilges, J. Markley and E. Laue (2002). The
CCPN project: An interim report on a data model for the NMR community
(Progress report). Nature Struct. Biol. 9, 416-418.

Rasmus H. Fogh, Wayne Boucher, Wim F. Vranken, Anne
Pajon, Tim J. Stevens, T.N. Bhat, John Westbrook, John M.C. Ionides and
Ernest D. Laue (2005). A framework for scientific data modeling and automated
software development. Bioinformatics 21, 1678-1684.

===========================REFERENCE END===============================

"""
# miscellaneous useful functions

import math, string, sys, types, os

def formatFloat(x, places = 3):
  """autoformat float to specified number of significant figures
  currently does not do scientific formatting at all
  """

  try:
    s = int(math.floor(math.log(abs(x)) / math.log(10)))
  except:
    s = 0
  d = max(0, places - s - 1)
  t = '%%.%sf' % d
  return t % x
 
# autoformat float to specified number of decimals
def formatDecimals(x, decimals = 0):
  """autoformat float to specified number of decimals
  """

  t = '%%.%df' % max(0, decimals)
  return t % x
 
def upperFirst(s):
  """uppercase first letter
  """
  return s[0].upper() + s[1:]

def lowerFirst(s):
  """lowercase first letter
  """
  return s[0].lower() + s[1:]

def capitalizeUnderscore(s):
  """capitalize entire string putting underscores in
  before existing capital letters (except first one)
  """

  t = ''
  for n in range(len(s)):

    if ((n > 0) and (s[n] in string.uppercase)):
      t = t + '_'

    t = t + s[n].upper()

  return t

def substituteName(string, name, n = 1):
  """substitute name n times in format string
  """

  return string % (n * (name,))

def toBoolean(x):
  """ Convert x to true/false. 
  Could be inlined, but function is more intelligible.
  """
  return x and True or False

def checkListIsSet(w):
  """ Checks if w (which could be list or tuple) is a set, i.e. no repeats.
  Creates a copy of the list or tuple to do the check.
  """
  z = list(w)
  for x in w:
    z.remove(x)
    assert x not in z, 'repeated element "%s"' % x

def isArray(x):
  """ Returns true if x is tuple or list, false otherwise.
  """

  if (type(x) in (types.TupleType, types.ListType)):
    return True
  else:
    return False

def isBigEndian():

  """ Returns true if platform is big endian, false if little endian.
  """

  if (sys.byteorder == 'big'):
    return True
  else:
    return False

def productArray(array):

  """ Returns product of entries of array
  """

  return reduce(lambda x, y: x*y, array, 1)

def cumulativeProductArray(array):

  """ Returns product of entries of array and also cumulative array of products
  """

  n = len(array)
  cumulative = n*[0]
  product = 1
  for i in range(n):
    cumulative[i] = product
    product = product * array[i]

  return (product, cumulative)

def indexOfArray(array, cumulative):

  index = 1
  for i in range(len(cumulative)):
    index = index + array[i] * cumulative[i]

  return index

def arrayOfIndex(index, cumulative):

  n = len(cumulative)
  array = n*[0]
  for i in range(n-1, -1, -1):
    array[i] = index / cumulative[i]
    index = index % cumulative[i]

  return array


def compactStringList(stringList, separator='', maxChars=80):
  """ compact stringList into shorter list of longer strings,
  each either made from a single start string, or no longer than maxChars
  
  From previous breakString function.
  Modified to speed up and add parameter defaults, Rasmus Fogh 28 Aug 2003
  Modified to split into two functions 
  and to add separator to end of each line, Rasmus Fogh 12 Sep 03
  Modified to separate string breaking from list modification
  Rasmus Fogh 29/6/06
  Modified to return single-element lists unchanged
  Rasmus Fogh 29/6/06
  """
  
  result = []
  
  if not stringList:
    return result
  elif len(stringList) ==1:
    return stringList[:]
  
  seplength = len(separator)
  
  nchars = len(stringList[0])
  start=0
  for n in range(1,len(stringList)):
    i = len(stringList[n])
    if nchars + i + (n-start)*seplength > maxChars:
      result.append(separator.join(stringList[start:n] + ['']))
      start = n
      nchars = i
    else:
      nchars = nchars + i
  result.append(separator.join(stringList[start:len(stringList)]))
  
  return result


def divideString(text, separator=' ', maxChars=72):
  """ divide string in series of substrings no longer than maxchars
  
      From previous breakString function.
      Modified to speed up and add parameter defaults, Rasmus Fogh 28 Aug 2003
      Modified to split into two functions
      and to add separator to end of each line, Rasmus Fogh 12 Sep 03
      Modified to separate string breaking from list modification
      Rasmus Fogh 29/6/06
      Added special case for text empty or None
      Rasmus Fogh 07/07/06
  """
  
  if not text:
    return ''
  
  return compactStringList(text.split(separator), separator=separator, 
                           maxChars=maxChars)

def breakString(text, separator=' ', joiner='\n', maxChars=72):
 
  """ Splits text on separator and then joins pieces back together using joiner
      so that each piece either single element or no longer than maxChars
      
      Modified to speed up and add parameter defaults, Rasmus Fogh 28 Aug 2003
      Modified to split into two functions 
      and to add separator to end of each line
      Modified to separate string breaking from list modification
      Rasmus Fogh 29/6/06
      Added special case for text empty or None
      Rasmus Fogh 07/07/06
  """
  
  if not text:
    return ''
  
  t = compactStringList(text.split(separator), separator=separator, 
                        maxChars=maxChars)
  
  return joiner.join(t)

def documentationFormat(text):
  """ Converts text to a multiline string, and returns a string literal
  expression with one line per line that evaluates to the multiline string
  Modified to separate string breaking from list modification
  Rasmus Fogh 29/6/06
  """

  if not text:
    return '""'
  
  ll = []
  
  for ss in text.splitlines(True):
    ll.extend(compactStringList(ss.split(' '), separator=' ',  maxChars=60))
  
  if len(ll) == 1:
    return repr(ll[0])
  
  if not ll[-1]:
    ll[-1] = '\n'
  elif ll[-1][-1] != '\n':
    ll[-1] = ll[-1] + '\n'
  
  return """(%s
)""" % '\n'.join(map(repr,ll))

def returnFloat(x,default = 0.0, verbose = True):

  """
  Returns a float, or default if fails
  Note that <number>e+n strings are converted OK!
  """

  try:
    x = float(x)
  except:
    if verbose:
      print("Error converting '" + str(x) + "' to float: set to {}".format(str(default)))
    x = default
  return x

def returnFloats(xlist, verbose = True):

  """
  Converts elements of a list to floats
  """

  for n in range(0,len(xlist)):
    xlist[n]=returnFloat(xlist[n], verbose = verbose)
  return xlist

def returnLong(x, default = 0.0, verbose = 1):

  """
  Returns a long, or default if fails
  """

  try:
    x = long(x)
  except:
    if verbose:
      print("Error converting '" + str(x) + "' to long: set to {}".format(str(default)))
    x = default
  return x

def returnLongs(xlist, verbose = True):

  """
  Converts elements of a list to longs
  """

  for n in range(0,len(xlist)):
    xlist[n]=returnLong(xlist[n],verbose = verbose)
  return xlist

def returnInt(x,default = 0, verbose = True):

  """
  Returns an int, or default if fails
  """

  try:
    x = int(x)
  except:
    if verbose:
      print("Error converting '" + str(x) + "' to integer: set to {}.".format(str(default)))
    
    x = default
      
  return x

def returnInts(xlist,verbose = True):

  """
  Converts elements of a list to ints
  """

  for n in range(0,len(xlist)):
    xlist[n]=returnInt(xlist[n], verbose = verbose)
  return xlist

def returnStrings(xlist):

  """
  Converts elements of a list to strings
  """

  newList = xlist[:]
  for n in range(0,len(newList)):
    newList[n]=str(newList[n])
  return newList

def returnList(value):
  
  """
  Returns a list from a single value or tuple
  """
  
  if type(value) == type(()):
    value = list(value)
  elif value and type(value) != type([]):
    value = [value]
  
  return value

def unquote(cols,quote):
  
  """
  Remove quote for each element in list
  """
  
  for n in range(0,len(cols)):
    if cols[n][0] == quote:
      cols[n] = cols[n][1:]
    if cols[n][-1] == quote:
      cols[n] = cols[n][0:-1]
  return cols

def joinUnquote(cols,quote,joinString = " "):
  
  """
  Remove quote
  """
  
  if cols[0][0] == quote:
    cols[0] = cols[0][1:]
  if cols[-1][-1] == quote:
    cols[-1] = cols[-1][0:-1]
  # Rejoin with single spaces
  return joinString.join(cols)

def drawBox(text,indent = "",liner = "#"):

  """
  Draw a box around a string with an indent and a selected liner
  """

  box = indent
  box += liner * (len(text)+4)
  box += os.linesep + indent + liner + " " + text + " " + liner + os.linesep + indent
  box += liner * (len(text)+4)
  box += os.linesep
  return box 
  
def getUpperPowerTwo(value):
  
  """
  Get next or current power of two
  
  NOTE: if value = power of two, then this function returns this value
  """
  
  for i in range(0,25):

    t = 2 ** i
    
    if t >= value:
      return t

  return None

def getPowerTwoExp(value):
  
  """
  Gives the exponent (2 ** n) that leads to the value
  (or the first power of two above this value)
  """
  
  for i in range(0,25):

    t = 2 ** i
    
    if t >= value:
      return i
  
  return None


def getMeanValue(valueList, valueNumber = None, valueTotal = None):
  
  """
  Get the mean value for the values in valueList with total
  valueTotal and number of values valueNumber (if given, else is calculated)
  """

  if not valueTotal:
    
    valueTotal = 0
    
    for value in valueList:
      valueTotal += value
      
  if not valueNumber:
    valueNumber = len(valueList)

  if valueNumber < 1:
    valueMean = None
  else:
    valueMean = (valueTotal * 1.0) / valueNumber
  
  return valueMean

def getStandardDev(valueList,valueTotal = None):

  """
  Get the standard deviation for the values in valueList with total
  valueTotal (if given, else is calculated)
  """
  
  valueNumber = len(valueList)
  valueAverage = getMeanValue(valueList,valueNumber = valueNumber, valueTotal = valueTotal)  
  
  if valueAverage == None:
    return None

  squaredSum = 0.0
  
  for value in valueList:
  
    squaredSum += (value - valueAverage) ** 2
  
  if valueNumber > 1:
    standardDev = (squaredSum/(valueNumber - 1)) ** 0.5
  else:
    standardDev = 0.0
  
  return standardDev

def getRms(valueList, total = None):
  
  """
  Calculates the root mean square for a list of values.
  """

  sqSum = 0.0
  
  for value in valueList:
    sqSum += value ** 2
    
  if not total:
    total = len(valueList)
  
  rms = (sqSum / total) ** 0.5
  
  return rms

def stringToList(listString):
  
  """
  Converts a str([]) back to []
  NOTE: still have to convert back to integer/float!
  """
  
  listString = listString[1:-1]
  
  newList = listString.split(',')
  
  return newList
  
  
def nameToSqlName (name):
  """
  Function that returns sql name from a python name.
  Changes anExampleString51 to AN_EXAMPLE_STRING_51
  """

  sqlName = ""
  previousChar = ''
  for char in name:
    if char.isupper():
      sqlChar = "_" + char.lower()
      sqlName = sqlName + sqlChar
    elif char.isdigit() and not previousChar.isdigit():
      sqlChar = "_" + char
      sqlName = sqlName + sqlChar
    else:
      sqlName = sqlName + char
    previousChar = char

  if sqlName[0] == "_":
    sqlName = sqlName[1:]
    
  return sqlName.upper()

def frange(start, end=None, inc=None):

  "A simple range function that accepts float increments. Taken from ASPN: http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/66472"

  if end == None:
    end = start + 0.0
    start = 0.0
            
  if inc == None:
    inc = 1.0

  L = []

  while 1:
    next = start + len(L) * inc
    if inc > 0 and next >= end:
      break
    elif inc < 0 and next <= end:
      break
    L.append(next)
                    
  return L
      
def makePowerSet(valueList):

  """
  Returns the powerset for a list.
  
  Use as: powersetlist = makePowerSet(inputList)
  
  For [1,2,3] or [1,3,2] both will return 
  [[], [1], [2], [1, 2], [3], [1, 3], [2, 3], [1, 2, 3]]
  
  """
  
  # copy (to avoid side effects) and sort
  values = list(valueList)
  values.sort()
  
  # calculate result
  result = [[]]
  for value in values:
    length = len(result)
    for ll in result[:length]:
      result.append(list(ll))
      result[-1].append(value)
  #
  return result


def semideepcopy(dd, doneDict=None):
  """ does a semi-deep copy of a nested dictionary, for copying mappings.
  Dictionaries are copied recursively, .
  Lists are copied, but not recursively.
  In either case a single copy is made from a single object 
  no matter how many times it appears.
  Keys and other values are passed unchanged
  """
  
  if doneDict is None:
    doneDict = {}
  
  key = id(dd)
  result = doneDict.get(key)
  if result is None:
    result = {}
    doneDict[key] = result
 
    for kk,val in dd.items():
 
      if type(val) == types.DictType:
        result[kk] = semideepcopy(val, doneDict)
 
      elif type(val) == types.ListType:
        key2 = id(val)
        newval = doneDict.get(key2)
        if newval is None:
          newval = val[:]
          doneDict[key2] = newval
 
        result[kk] = newval
 
      else:
        result[kk] = val
  #
  return result

def isWindowsOS():

  return sys.platform[:3].lower() == 'win'

def isMacOS():

  return sys.platform.lower() == 'darwin'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import setuptools
import os

parent_path = os.path.relpath(".")
path_to_file = os.path.join(parent_path, "README.md")

with open(path_to_file, "r") as f:
    long_description = f.read()

setuptools.setup(
    name="b2bTools",
    version="3.0.5b3",
    author="Wim Vranken",
    author_email="Wim.Vranken@vub.be",
    description="bio2Byte software suite to predict protein biophysical properties from their amino-acid sequences",
    license="OSI Approved :: GNU General Public License v3 (GPLv3)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    maintainer="Jose Gavalda-Garcia, Adrian Diaz, Wim Vranken",
    maintainer_email="jose.gavalda.garcia@vub.be, adrian.diaz@vub.be, wim.vranken@vub.be",
    url="https://bio2byte.be",
    project_urls={
        # "Bug Tracker": "https://bitbucket.org/bio2byte/b2btools/issues",
        "Documentation": "https://bio2byte.be/b2btools/package-documentation",
        # "Source": "https://bitbucket.org/bio2byte/b2btools/",
        "HTML interface" : "https://bio2byte.be/b2btools"
    },
    packages=setuptools.find_packages(exclude=("**/test/**",)),
    include_package_data=True,
    keywords="b2bTools,biology,bioinformatics,bio-informatics,fasta,proteins,protein-folding",
    classifiers=[
        "Natural Language :: English",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Scientific/Engineering :: Physics",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Education",
        "Development Status :: 5 - Production/Stable"
    ],

    python_requires=">=3.6, <3.10",

    install_requires=[
        "joblib",
        "networkx",
        "pomegranate",
        "numpy",
        "Pillow",
        "PyYAML",
        "scikit-learn>=1.0.1",
        "scipy",
        "six",
        "threadpoolctl",
        "torch>=1.8.0",
        "torchvision",
        "pandas",
        "requests",
        "biopython",
        "matplotlib"
    ]
)



# File to handle different inputs and run the predictors (single seq level)
import os, os.path, sys
import json
import argparse
import numpy as np
import pandas as pd
from Bio.PDB import *
import urllib.request
from b2bTools.general.Io import B2bIo
from b2bTools.singleSeq.Predictor import MineSuite
from b2bTools.general.plotter import plot_results

# Input files parser
parser = argparse.ArgumentParser(description='Input handler')

parser.add_argument('-pdb', action='store', nargs = '*',
                    help='Use pdb id or file as input to get the sequence and \
                          run the predictors')

parser.add_argument('-uniprot', type=str,
                    help='Uniprot ID code')

parser.add_argument('-fasta', type=str,
                    help='The input protein sequence/s in fasta format')

parser.add_argument('-nef', type=str,
                    help='The input protein file in nef format')

args = parser.parse_args()

aa_coder = {'CYS': 'C', 'ASP': 'D', 'SER': 'S', 'GLN': 'Q', 'LYS': 'K',
     'ILE': 'I', 'PRO': 'P', 'THR': 'T', 'PHE': 'F', 'ASN': 'N',
     'GLY': 'G', 'HIS': 'H', 'LEU': 'L', 'ARG': 'R', 'TRP': 'W',
     'ALA': 'A', 'VAL':'V', 'GLU': 'E', 'TYR': 'Y', 'MET': 'M'}

###############################################################################
###############################################################################

def run_predictors(seqs, id):
    tools = B2bIo()
    ms = MineSuite()
    ms.predictSeqs(seqs)

    with open("tmp_file.json", 'w') as fp:
        json_data = ms.getAllPredictionsJson(id)
        json.dump(json_data, fp)
        input_csv = input("Do you want to save the predictions in a text file? (Y/N)")

        if input_csv == "Y":
            tools.json_preds_to_csv_singleseq(json_data)
            filename = json_data["id"].split("/")[-1].split(".")[0]
            print("Results succesfully saved in results/{}.txt".format(filename))

    plot_results("tmp_file.json", msa_like=1,
                 mutation="")


###############################################################################
###############################################################################

if __name__ == '__main__':
    if args.pdb:
        # Iput ID: The two argument values must be the pdb Id and the chain Id
        if len(args.pdb) == 2:
            pdb_id_input = args.pdb[0]
            chain_id_input = args.pdb[1]
            print ("PDB code and chain:", pdb_id_input, chain_id_input)
            full_pdb_id = [pdb_id_input.upper(), chain_id_input]
            seqs = []
            if len(full_pdb_id) > 1 and len(full_pdb_id[0]) == 4:
                try:
                    url = 'https://www.rcsb.org/fasta/chain/'
                    pdburl = url + ".".join(full_pdb_id) + '/download'
                    response = urllib.request.urlopen(pdburl)
                    data = response.read()
                    PDB_seq = data.decode('utf-8')
                    seqs.append((full_pdb_id[0], PDB_seq.split("\n")[1]))
                    run_predictors(seqs, full_pdb_id[0])

                except:
                    print("Entry not found / some error occured")

            else:
                print("Invalid entry check your input")

        # Input pdb file: The argument value must be the path to the pdb file
        if len(args.pdb) == 1:
            print("TODO")
            # Add warning for pdb
            # get sequence from pdb file

    elif args.uniprot:
        uniprotID = args.uniprot
        run_predictors(B2bIo.retrieve_seq_uniprot(uniprotID), uniprotID)

    elif args.fasta:
        tools = B2bIo()
        seqs = tools.readFasta(args.fasta)
        run_predictors(seqs, args.fasta)

    elif args.nef:
        tools = B2bIo()
        info = tools.readNefFileSequenceShifts(args.nef)
        sequence = [(args.nef ,''.join([aa_coder[resname[0]["residue_name"]] for
                                        resname in info["A"]]))]
        run_predictors(sequence, args.nef)



import os, os.path, sys
import requests
import time
import urllib.parse
import urllib.request
import json
import argparse
import numpy as np
import pandas as pd
from Bio import SeqIO
from Bio.Blast import NCBIWWW, NCBIXML
from b2bTools.general.Io import B2bIo
from b2bTools.multipleSeq.Predictor import MineSuiteMSA
from b2bTools.singleSeq.Predictor import MineSuite
from b2bTools.general.plotter import Plotter

###############################################################################
###############################################################################

class UniRef50Manager:
    def __init__(
            self,
            UniProtKB_idA,
            fasta_name_A = None,
            ):
        # UniProtKb IDs
        self.idA = self._retrieve_id(UniProtKB_idA)
        # UniRef50 IDs
        self.UniRef50_idA = self._retrieve_UniRef50_id(self.idA)

        # UniRef50 members and protein sequences
        self.UniRef50_membersA_fasta = \
            self._retrieve_UniRef50_members_fasta(
                    self.UniRef50_idA,
                    fileName=fasta_name_A,
                    )

    def UniProtKB_Mapping(self,
            query,
            fileName=None,
            From="ACC",
            To="ACC",
            Format="fasta",
            Columns="",
            limit="10"
    ):
        """Implementation of the UniProtKB mapping REST api.

        UniProtKB provides an REST API to map one database to another,
        similar to the webservice (https://www.uniprot.org/uploadlists/).
        More information about the working of the API can be found on:
        https://www.uniprot.org/help/api_idmapping

        Parameters
        ----------
        query : str
            The query parameter accepts a sequence of identifiers separated
            by a space or newline character.  The format of the identifiers
            depends on the database which is mapped from.
        fileName : str (Default=None)
            The destination path of the file that is downloaded.  If just a
            name is specified, the file is stored in the directory from
            which the function was executed.  Additionaly the relative or
            absolute path can be specified.
        From : str (Default='ACC')
            The database from which the identifiers are mapped. The
            available databases are listed on:
            https://www.uniprot.org/help/api_idmapping
        To : str (Default='ACC')
            The database to which the identifiers are mapped. The
            available databases are listed on:
            https://www.uniprot.org/help/api_idmapping
        Format : str (Default='fasta')
            The file format of the mapped identifiers.  The available format
            are:
            html | tab | xls | fasta | gff | txt | xml | rdf | list | rss
        Columns : str (Default="")
            The column information to be downloaded for each entry when the
            format is tab or xls.

        Returns
        -------
        response : str
            The content of the downloaded file in string format
        """

        """ The Download of the file is tried 10 times.  Each time the
        function waits a bit longer. Sometimes remote servers are busy and
        repeating a failed request helps.
        """
        for i in range(10):
            try:
                url = 'https://www.uniprot.org/uploadlists/'
                params = {
                    "query": query,
                    "from": From,
                    "to": To,
                    "format": Format,
                    "columns": Columns,
                    "limit": limit
                }
                data = urllib.parse.urlencode(params)
                data = data.encode('utf-8')
                req = urllib.request.Request(url, data)
                with urllib.request.urlopen(req) as f:
                    response = str(f.read(), encoding="utf-8")

                if fileName:
                    # Write content to file
                    with open(fileName, "w") as f:
                        # To limit the number of hits (Otherwise align is too slow)
                        response = "\n>".join(response.split("\n>")[:25])
                        f.write(response)
                return response
            except:
                print(
                    "request failed, wait for",
                    i * 5,
                    "seconds and try again",
                )
                time.sleep(i * 5)


    def _retrieve_id(
            self,
            UniProtKB_id,
            ):
        """ Retrieve the UniProtKB id.

        Sometimes, UniProtKB updates identifiers.  This method maps the
        a UniProtKB identifier to its own database to retrieve the
        latest version of this identifier.

        Parameters
        ----------
        UniProtKB_id : str
            The UniProtKB_id identifier (can be outdated).

        Returns
        -------
        new_UniProtKB_id : str
            When the UniProtKB_id is outdated, the up to date identifier
            is returned instead.  Otherwise, the original identifier
            is returned.
        """
        new_UniProtKB_id = self.UniProtKB_Mapping(
                UniProtKB_id,
                Format="list",
                ).strip()
        return new_UniProtKB_id

    def _retrieve_UniRef50_id(
            self,
            UniProtKB_id,
            ):
        """ Retrieve the UniRef50 identifier.

        UniProtKB clusters all of the protein sequences by sequence
        identity, called UniRef groups.  All members of a UniRef50 group
        will have at most 50 percent sequence identity.

        Parameters
        ----------
        UniProtKB_id : str
            The UniProtKB_id identifier (can be outdated).

        Returns
        -------
        UniRef50_id : str
            The UniRef50 identifier.
        """
        UniRef50_id = self.UniProtKB_Mapping(
                UniProtKB_id,
                From="ACC",
                To="NF50",
                Format="list",
                ).strip()
        return UniRef50_id

    def _parse_fasta(
            self,
            fasta_string,
            ):
        """ Parse fasta string into dictionary.

        This function parses the fasta file (in string format) into a
        dictionary of protein sequences.

        Parameters
        ----------
        fasta_string : str
            A fasta file in string format.

        Returns
        -------
        fasta_dictionary : dict
            The fasta file in dictionary format.
                key : ID
                value : protein sequence
        """
        fasta_dictionary = dict()
        fasta_lines = (line for line in fasta_string.split("\n"))
        for line in fasta_lines:
            if line.startswith(">"):
                Id = line.replace(">","")\
                    .split("|")[1]
                fasta_dictionary[Id] = ""
            else:
                fasta_dictionary[Id] += line.strip()
        return fasta_dictionary

    def _retrieve_UniRef50_members_fasta(self, UniRef50_id, fileName=None):
        """Retrieve fasta file of UniRef50 members.

        All members of an UniRef50 cluster have a sequence identity of
        50% or more.  This means they are homologues to each other.
        The UniProtKB mapping API is used to retrieve those members in
        fasta format.  The resulting string is subsequently parsed to a
        dictionary.

        Parameters
        ----------
        UniRef50_id : str
            The UniRef50 identifier.
        fileName : str
            Name of the fasta file.

        Returns
        -------
        UniRef50_members_fasta : dict
            Returns a dictionary of the members and their protein
            sequence.
                key : UniProtKB ID
                value : protein sequence
        """
        # if no fileName give, create generic one
        if not fileName:
            fileName = "{}.fasta".format(UniRef50_id)
        # Retrieve fasta string with UniProtKB mapping service
        UniRef50_members_fasta_str = self.UniProtKB_Mapping(
                UniRef50_id,
                From="NF50",
                To="ACC",
                Format="fasta",
                fileName=fileName
                )

        # Parse string into dictionary of protein sequences
        # UniRef50_members_fasta = self._parse_fasta(UniRef50_members_fasta_str)

        # Run the MSA using all the seqs in the fasta file found
        self.aligned_file = '{}_tcof.aln'.format(fileName.split('.')[0])
        cmd = "t_coffee {} -output=fasta_aln -outfile={}".format(
        fileName, self.aligned_file)
        os.system(cmd)

###############################################################################
###############################################################################

class BlastManager:
    def __init__(self, seq, file_name = None, mut_option = None, mut_position = None, mut_residue = None):
        self.rrm_seq = seq
        self.target_seq = []
        with open(self.rrm_seq, 'r') as f:
            for rec in SeqIO.parse(f, 'fasta'):
                self.target_seq = [res for res in rec.seq]
                break  # If the fasta file provided have more than one seq it
                # will only process the first one

        self.target_seq = ''.join(self.target_seq)

        if file_name:
            self.file_name = file_name
        else:
            self.file_name = str(
                input('Write the output filename for the BLAST results'
                    ' (String without extension nor spaces):\n'
                    '>>>') + '_blast.xml')
        if not mut_option:
            mut_option = input('Do you want to generate a mutated sequence?:\n'
                '(y = Yes; n = No) >>>')

        if mut_option == 'y':
            self.mutated_seq = self.mutator(self.target_seq, mut_position=mut_position, mut_residue=mut_residue)

        else:
            self.mutated_seq = None

    def mutator(self, prot_seq, mut_position = None, mut_residue = None):
        #     Mutagenicity test, mutate the target sequence and add it
        if mut_position:
            mut_pos = mut_position
        else:
            mut_pos = int(
                input('Select one position in the sequence to mutate:\n' +
                    prot_seq + '\n>>>'))

        print('You have selected {}{} to mutate'.format(
            prot_seq[mut_pos], mut_pos))

        if mut_residue:
            mut_res = mut_residue
        else:
            mut_res = str(input('Select the residue to mutate to:\n' +
                                'A, C, D, E, F, G, H, I, K, L, M, N, P, Q, '
                                'R, S, T, V, W, Y' +
                                '\n>>>'))

        self.mutation = prot_seq[mut_pos] + str(mut_pos) + mut_res

        self.mutated_seq = list(prot_seq)
        self.mutated_seq[mut_pos] = mut_res
        return(self.mutated_seq)

    # Function to run the online-based version of BLAST (qblast) and generate
    # the output xml file
    def run_qblast(self, blast_prog='blastp', db='refseq_protein'):
        fasta_string = open(self.rrm_seq).read()
        result_handle = NCBIWWW.qblast(blast_prog, db, fasta_string,
                                       format_type='XML')

        with open("{}".format(self.file_name), "w") as out_handle:
            out_handle.write(result_handle.read())

        self.parse_blast(self.file_name)

    # Parse the xml file with blast results generated in te function above
    def parse_blast(self, file_name):
        print('Running BLAST')
        # Debug line to input the xml file manually instead of running qblast
        result_handle = open(file_name)
        blast_record = NCBIXML.read(result_handle)

        # Define the column names that we wanna have in the final table
        column_names = ["accesion_version", "sequence_aligned",
                        "alignment_length", "identity", "coverage", "e_value"]
        df = pd.DataFrame(columns=column_names)

        # Parse the aligment file by calculating the coverage and identities
        # and filter the results according to those and the e-value as well
        self.filepath = '{}.txt'.format(
            self.file_name.split('.')[0])

        print(self.filepath)
        with open(self.filepath, 'w+') as the_file:
            # If the file is empty append first the target & mutated sequence
            if os.stat(self.filepath).st_size == 0:
                the_file.write('>target_sequence\n')
                the_file.write(self.target_seq + '\n')
                if self.mutated_seq:
                    the_file.write('>mutated_sequence\n')
                    the_file.write(''.join(self.mutated_seq) + '\n')

            # Keep blast results with ID > 0.9, cov > 0.9, e-value < 0.04
            for alignment in blast_record.alignments:
                for hsp in alignment.hsps:
                    coverage = hsp.align_length / blast_record.query_length
                    identity = hsp.identities / hsp.align_length
                    if identity > 0.9 and coverage > 0.9 and hsp.expect < 0.04:
                        acc_version = alignment.title.split("|")[1]
                        the_file.write('>{}\n'.format(acc_version))
                        the_file.write(
                            '{}'.format(hsp.sbjct.replace('-', '')) +
                            '\n')

        # Run the MSA using all the BLAST matches found
        self.aligned_file = '{}_tcof.aln'.format(self.file_name.split('.')[0])
        cmd = "t_coffee {} -output=fasta_aln -outfile={}".format(
            self.filepath, self.aligned_file)
        os.system(cmd)


###############################################################################
###############################################################################

class MsaManager:
    def __init__(self):
        print('Running MsaManager')

    # Function that uses t-coffee software to align two MSAs
    def t_coffe_MSA_aligner(self, msa_files):
        if os.path.exists('msa_tmp.aln'):
            os.remove('msa_tmp.aln')
        self.out_file_name = 'msa_tmp.aln'
        self.msa_dict = {}

        for file in msa_files:
            filename = file.split('/')[-1].split('.')[0]
            self.msa_dict[filename] = []
            f = open(file, "r")
            self.msa_dict[filename] = [l.split('>')[1].split('\n')[0] for
                                       l in f if l.startswith('>')]

        msa_files_tcof = ','.join(msa_files)
        cmd = "t_coffee -quiet -profile={} -outfile={}".format(msa_files_tcof,
                                                        self.out_file_name)
        os.system(cmd)


###############################################################################
###############################################################################

class predManager:
    def __init__(self):
        print("Biophysical predictor class running...")
        self.ms = MineSuiteMSA()

    # Function to run all the predictors over an MSA file:
    # DynaMine, EFoldMine and Disomine)
    def run_predictors(self, msa_file, predTypes = [], msa_map = None, mutation=None):
        # msa_map is a dictionary containing the original alignment with
        # file name (key) and sequence names (value), required when an
        # alignment of 2 MSAs is provided (None by default)
        self.ms.predictSeqsFromMSA(msa_file, predTypes=predTypes)

        self.mutated_seq_preds = None

        # Same but now mapped to reference sequence ID in MSA, gaps are None
        self.ms.predictAndMapSeqsFromMSA(msa_file, dataRead=True)
        self.allAlignedPredictions = self.ms.allAlignedPredictions

        self.jsondata_list = []

        try:
            # Create the results folder if it's not there yet
            if not os.path.exists("/results"):
                os.mkdir("/results")
        except BaseException:
            pass

        # When we work with an alignment of MSAs
        if msa_map != None:
            for key in msa_map.keys():
                # The two variables below are updated to include only the
                # sequences from one of the source MSA files at a time
                self.ms.allAlignedPredictions = {
                    sel_key: self.allAlignedPredictions[sel_key] for sel_key
                    in msa_map[key]}
                self.ms.allSeqIds = [
                    key for key in self.ms.allAlignedPredictions.keys()]

                # Same but now mapped to full MSA, distributions of scores.
                self.jsondata_list.append(self.ms.getDistributions())

        # When we work with a single MSA
        else:
            # Same but now mapped to full MSA, distributions of scores.
            self.jsondata_list.append(self.ms.getDistributions())

    def run_predictors_json_input(self, json_file, predTypes = [], output_dir=""):
        msa_file = B2bIo.json_to_fasta_variants(json_file, output_dir)
        print(predTypes)
        self.ms.predictSeqsFromMSA(msa_file, predTypes=predTypes)

        # Same but now mapped to full MSA, gaps are None
        self.ms.predictAndMapSeqsFromMSA(msa_file, dataRead=True)
        self.allAlignedPredictions = self.ms.allAlignedPredictions

###############################################################################
###############################################################################

if __name__ == '__main__':
    # Input files parser
    parser = argparse.ArgumentParser(description='Input handler')

    parser.add_argument('-aligned_file', type = str, nargs=1,
                        help = 'Runs the predictors over an aligned file')

    parser.add_argument('-qblast', type=str,
                        help='The input protein sequence/s in fasta format'
                            ' to run qBLAST and the biophysical predictions')

    parser.add_argument('-align', type=str, nargs=2,
                        help='MSA files to be aligned and biophysically compared')

    parser.add_argument('-uniref', type=str, nargs=1,
                        help='UniprotKB id to find related sequences and compare them biophysically')

    parser.add_argument('-json', type=str, nargs=1,
                        help='JSON file with the WT sequence and the variants defined')

    args = parser.parse_args()

    if args.aligned_file:
        print(args.aligned_file)
        PRED = predManager()
        PRED.run_predictors(args.aligned_file[0])

    # Run blastp with fasta file as input
    elif args.qblast:
        RRM = BlastManager(args.qblast)
        print("Running BLAST (This step may take a few minutes)")
        RRM.run_qblast()
        PRED = predManager()
        PRED.run_predictors(RRM.aligned_file, mutation = RRM.mutation)

    # Align two MSA files, run the predictors and plot the mapped results
    elif args.align:
        MSA = MsaManager()
        MSA.t_coffe_MSA_aligner(args.align)
        PRED = predManager()
        PRED.run_predictors(MSA.out_file_name, msa_map = MSA.msa_dict)

    # UniprotKB id to find related sequences and compare them biophysically
    elif args.uniref:
        print(args.uniref)
        uniref = UniRef50Manager(args.uniref[0], 'uniref')
        PRED = predManager()
        PRED.run_predictors(uniref.aligned_file)

    # To define several sequences and variant through a json file format
    elif args.json:
        print(args.json)
        PRED = predManager()
        PRED.run_predictors_json_input(args.json[0])

    else:
        print("Not argument provided or invalid argument")



import numpy as np
import time

from b2bTools.singleSeq.Predictor import MineSuite
from b2bTools.general.Io import B2bIo

class MineSuiteMSA(MineSuite):

  scriptName = "b2bTools.msaBased.Predictor"

  def predictSeqsFromMSA(self, msaFile, gapCode='-', predTypes=[]):

    # This will read in alignment, should automatically detect format. Code is in general/Io.py
    self.seqAlignments = self.readAlignments(msaFile, resetAlignRefSeqID=True, gapCode=gapCode)

    seqs = []
    for seqId in self.seqAlignments.keys():
      seqs.append((seqId,self.seqAlignments[seqId].replace(gapCode,'')))

    self.predictSeqs(seqs, predTypes= predTypes)

    # Now self.allPredictions will give you the predictions for all the individual sequences in the MSA!

  def predictAndMapSeqsFromMSA(self, msaFile, gapCode='-', dataRead=False,
                               predTypes=[]):

    # Read in data only if not yet present - can re-use this function within instance of class if data already present!
    if not dataRead:
      self.predictSeqsFromMSA(msaFile,gapCode=gapCode, predTypes=predTypes)

    self.allSeqIds = list(self.seqAlignments.keys())
    self.allSeqIds.sort()

    # All the current prediction types
    self.predictionTypes = self.allPredictions[self.allSeqIds[0]].keys()

    self.allAlignedPredictions = {}
    sequenceInfo = {}

    for seqId in self.allSeqIds:

      alignment = self.seqAlignments[seqId]
      seqIndex = 0

      sequenceInfo[seqId] = []
      self.allAlignedPredictions[seqId] = {}

      for predictionType in self.predictionTypes:
        self.allAlignedPredictions[seqId][predictionType] = []

      for alignIndex in range(len(alignment)):
        if alignment[alignIndex] == self.gapCode:
          for predictionType in self.predictionTypes:
            self.allAlignedPredictions[seqId][predictionType].append(None)
        else:
          resName = self.allPredictions[seqId]['backbone'][seqIndex][0]

          assert resName == alignment[alignIndex] or resName == 'X', "Amino acid code mismatch {}-{}".format(resName,
                                                                                                             alignment[
                                                                                                               alignIndex])

          sequenceInfo[seqId].append(alignment[alignIndex])

          for predictionType in self.predictionTypes:
            try:
              valueTuple = self.allPredictions[seqId][predictionType][seqIndex]
              resName, predValue, *other_values = valueTuple
              self.allAlignedPredictions[seqId][predictionType].append(predValue)
            except ValueError:
              raise ValueError("Predicted value of type '%s' not found for input sequence '%s'" % (predictionType, seqId))

          seqIndex += 1

    self.allAlignedPredictions['sequence'] = sequenceInfo

  def filterByRefSeq(self,refSeqId):

    assert refSeqId in self.allSeqIds,  'Reference sequence ID {} missing in current prediction information!'.format(refSeqId)

    # Here filter so that get back values in reference to the sequence ID that is given.

  def getDistributions(self):

    distribKeys = ('median', 'thirdQuartile', 'firstQuartile', 'topOutlier', 'bottomOutlier')
    numDistribKeys = len(distribKeys)

    # Now generate the info for quartiles, ... based on the alignRefSeqID, first entry in alignment file
    self.alignedPredictionDistribs = {}
    for predictionType in self.predictionTypes:
      self.alignedPredictionDistribs[predictionType] = {}
      for distribKey in distribKeys:
        self.alignedPredictionDistribs[predictionType][distribKey] = []

    # Loop over whole alignment
    alignmentLength = len(self.seqAlignments[self.allSeqIds[0]])

    for alignIndex in range(alignmentLength):
      for predictionType in self.predictionTypes:
        predValues = [self.allAlignedPredictions[seqId][predictionType][alignIndex] for seqId in self.allSeqIds if
                      self.allAlignedPredictions[seqId][predictionType][alignIndex] != None]

        distribInfo = self.getDistribInfo(predValues)

        for i in range(numDistribKeys):
          self.alignedPredictionDistribs[predictionType][distribKeys[i]].append(distribInfo[i])

    self.jsonData = B2bIo.getAllPredictionsJson_msa(self, results = self.alignedPredictionDistribs)
    return self.jsonData

  def getDistribInfo(self,valueList, outlierConstant = 1.5):
    # JR: I put this try-except cause in some MSA we may have only one sequence
    # so the distribution values cannot be calculated
    try:
      median = np.median(valueList)
      upper_quartile = np.percentile(valueList, 75)
      lower_quartile = np.percentile(valueList, 25)
      IQR = (upper_quartile - lower_quartile) * outlierConstant

      return (median, upper_quartile, lower_quartile,upper_quartile + IQR,lower_quartile - IQR)

    except:
      return (None, None, None, None, None)



import os, warnings

hmmbuild_bin = 'hmmbuild'
hmmscan_bin = 'hmmsearch'

try:
  from b2bTools.singleSeq.PSPer.localConstants import hmmbuild_bin
  from b2bTools.singleSeq.PSPer.localConstants import hmmscan_bin
except:
  pass

# NOTE: this will not work on Windows, fix this to be more generic
hmmbuildAbsent = os.system('which {}'.format(hmmbuild_bin))

if hmmbuildAbsent:
  warnings.warn("""
PSPer can only run if HMMER is installed.
Please install it from http://hmmer.org on your computer, and refer to this installation in the
   b2bTools/python/b2bTools/singleSeq/PSPer/Constant.py
file by pointing the hmmbuild_bin and hmmscan_bin variables at the binaries.""")


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  shiftcrypt.py
#
#  Copyright 2019  <orlando.gabriele89@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
#
import warnings
warnings.filterwarnings("ignore")
import torch
import argparse, subprocess, os, errno
from b2bTools.singleSeq.PSPer.source.utils import leggifasta
from b2bTools.singleSeq.PSPer.phase_transition_hmm import phase_hmm

def run_psper():

	pa = argparse.ArgumentParser()

	pa.add_argument('-i', '--infile',
						help='the input fasta file',
						required=True,
						)
	pa.add_argument('-viterbi', '--viterbi',
						action='store_true',
						help='provides the viterbi output',)
	pa.add_argument('-o', '--outfile',
						help='output file',
						default=None)
	args = pa.parse_args()
	if args.infile:
		prot=leggifasta(args.infile)
		try:
			prot=leggifasta(args.infile)
		except:
			print ('error in the parsing of the file. Please double check the fasta format. If everyhting is correct, please report the bug to orlando.gabriele89@gmail.com')
			return

		model=phase_hmm()
		model.fit()
		try:
			print ('running prediction')
			out=model.predict_proba(prot)
			if args.viterbi:
				print ('STARTING VITERBI')
				vit=model.viterbi(prot)
		except:
			raise
			print ('Error in the prediction phase. Please double check you installed all the dependencies. If everyhting is correct, please report the bug to orlando.gabriele89@gmail.com')
			return

		if args.outfile!=None:
			f=open(args.outfile,'w')
			for i in prot.keys():
				f.write(i+' '+str(out[i])+'\n')
				if args.viterbi:
					for k in range(len(prot[i])):
						f.write('\t'+prot[i][k]+' '+str(vit[i][k].split('_')[0])+'\n')
			f.close()
			#print out
		else:
			for i in prot.keys():
				print(i+' '+str(out[i])+'\n')
				if args.viterbi:
					for k in range(len(prot[i])):
						print('\t'+prot[i][k]+' '+str(vit[i][k].split('_')[0]))
			#print out
		print ('\nDone')
	else:
		print('infile argument not found')

def main(args):
		return 0

if __name__ == '__main__':
    run_psper()
